#!/usr/bin/python
"""
Auto3DEM script to change map orientation and particle parameters
from icosahedral (532/I90) to a different symmetry convention,
and viceversa. For now only the convertion to bring a 5-fold
vertex along the z-axis is supported. Specifically, the target
vertex is on the xy plane, -x/y quadrant. The rotation brings the
vertex to point towards the negative Z-axis.
"""
import sys, os
from glob import glob
from math import atan, sqrt, pi, sin, cos, radians, degrees, acos, atan2
import numpy as np

from pya3dm import ParamFile
from pya3dm import ConfigFile

try:
	from optparse import OptionParser
except:
	print "Module optparse not found: you need Python version 2.3 or later"
	sys.exit(-1)

    
def main():
    (options,args) =  parse_command_line()


# euler angles in radians
    golden_angle = atan(0.5*(1+sqrt(5)) - 1)
    (phi, theta, omega) = (-golden_angle, 0.5*pi, 0.) # rotation to bring vertex on -Z axis


    r = rotation_matrix_from_euler(phi, theta, omega)


# Detect type of input provided
    if istextfile(args[0]):
            if isconfigfile(args[0]):
                print("\nInput %s detected as configuration file.\n"%args[0])  
                cfgParams = ConfigFile(args[0])
                outcfgParams = cfg_transform(cfgParams, options.suffix, r, options.inverse_rot)
                if not options.inverse_rot:
                    if outcfgParams.hasParam('auto','symm_code'):
                        outcfgParams.setParam('auto','symm_code','1')
                    else:
                        outcfgParams.setParam('po2r','symm_code','1')
                        outcfgParams.setParam('p3dr','symm_code','1')
                    outcfgParams.Write(args[1])
            
            else:
                print("\nInput %s detected as stack parameter file.\n"%args[0])    
                datParams = ParamFile(args[0])
                outdatParams = dat_transform(datParams, r, options.inverse_rot)
                outdatParams.Write(args[1])
    else:
        print("\nInput %s detected as map file.\n"%args[0])    
        map_rotate(args[0], args[1], r, options.inverse_rot)


def dat_transform(indatparam, r, invrot):

    for p in indatparam.AllParticles():
        psi = radians(90 - p.omega)
        theta = radians(p.theta)
        phi = radians(p.phi)
        r0 = rotation_matrix_from_euler(phi, theta, psi)
        if invrot:
            r1 = r
        else:
            r1 = np.transpose(r)
        r2 = np.dot(r0, r1)
        (phi, theta, psi) = euler_from_rotation_matrix(r2)
        p.theta = degrees(theta)
        p.phi = degrees(phi)
        p.omega = degrees(0.5*pi - psi)
        
    return indatparam


def cfg_transform(incfgparam, sfx, r, invrot):

    rdir = os.path.join(os.path.dirname(incfgparam.cfgFilename),incfgparam.getParam('auto','rundir'))
    maps_transformed = []
    for m in ('start_map', 'start_map_even', 'start_map_odd'):
        if incfgparam.hasParam('auto',m):
            inmapcfg = incfgparam.getParam('auto',m)
            inmap = os.path.join(rdir,inmapcfg)
            if os.path.exists(inmap):
                if (invrot and os.path.splitext(inmapcfg)[0].endswith(sfx)):
                    outmapcfg = os.path.splitext(inmapcfg)[0][:-len(sfx)]+os.path.splitext(inmapcfg)[1]
                else:
                    outmapcfg = os.path.splitext(inmapcfg)[0]+sfx+os.path.splitext(inmapcfg)[1]
                outmap = os.path.join(rdir,outmapcfg)
                if not inmapcfg in maps_transformed:
                    map_rotate(inmap, outmap, r, invrot)
                    maps_transformed.append(inmapcfg)
                incfgparam.setParam('auto',m,outmapcfg)
    
    outdata = []
    for ind in incfgparam.getData():
        datParams = ParamFile(os.path.join(rdir,ind))
        outdatParams = dat_transform(datParams, r, invrot)
        if (invrot and os.path.splitext(ind)[0].endswith(sfx)):
            outd = os.path.splitext(ind)[0][:-len(sfx)]+os.path.splitext(ind)[1]
        else:
            outd = os.path.splitext(ind)[0]+sfx+os.path.splitext(ind)[1]
        outdata.append(outd)
        outdatParams.Write(os.path.join(rdir,outd))
        
    incfgparam.resetData()
    incfgparam.setData(outdata)
    
    return incfgparam



def map_rotate(inmap, outmap, r, invrot):

    if invrot: r = np.transpose(r)
    
# input rotation matrix
# write the transposed matrix
    rotfile = "_tmp_rotmat.txt"

    f = file(rotfile, "w")
    f.write("\n")
    f.write("%f, %f, %f\n"%(r[0][0],r[0][1],r[0][2]))
    f.write("%f, %f, %f\n"%(r[1][0],r[1][1],r[1][2]))
    f.write("%f, %f, %f\n"%(r[2][0],r[2][1],r[2][2]))
    f.write("%f, %f, %f\n"%(0., 0., 0.))
    f.close()
    
# input file for rot_map
    input_file = "_a3dm_icosrotate_input"
    f = file(input_file,"w")
    f.write("%s\n"%inmap)
    f.write("%s\n"%outmap)
    f.write("%s\n"%rotfile)
    f.close()

# launch program
    os.system("rot_map < %s"%input_file)
    
# cleanup
    os.remove(input_file)
    os.remove(rotfile)

    return


def rotation_matrix_from_euler(phi, theta, omega):
    """Generate a rotation matrix from Euler angle (z/y/z  convention).
    """

    # Trig.
    sint = sin(theta)
    cost = cos(theta)
    sinp = sin(phi)
    cosp = cos(phi)
    sino = sin(omega)
    coso = cos(omega)

    R = np.identity(3)
    
    R[0, 0] =   cosp * cost * coso - sinp * sino
    R[1, 0] =  -sino * cost * cosp - coso * sinp
    R[2, 0] =   sint * cosp

    R[0, 1] =   coso * cost * sinp + sino * cosp 
    R[1, 1] =  -sino * cost * sinp + coso * cosp
    R[2, 1] =   sinp * sint

    R[0, 2] =  -coso * sint
    R[1, 2] =   sint * sino
    R[2, 2] =   cost

    return R

def euler_from_rotation_matrix(R):
    """Generate Euler angles (z/y/z convention) from a rotation matrix.
    """
    theta = acos(R[2][2])
    if theta != 0. and theta != pi:
        phi = atan2(R[2][1],R[2][0])
        psi = atan2(R[1][2],-R[0][2])
    else:
        phi = 0.
        psi = atan2(-R[1][0],R[0][0])

    return phi, theta, psi

def isconfigfile(filename):
    from pya3dm.cfgparam import identifiers
    f = file(filename).readlines()
    for l in f:
        line = l.strip()
        if line and not line.startswith('#'):
            l1 = line.split()[0]
            if l1 in identifiers:
                return True
    return False

        
def istextfile(filename, blocksize = 512):
    return istext(open(filename).read(blocksize))

def istext(s):

    import string
    text_characters = "".join(map(chr, range(32, 127)) + list("\n\r\t\b"))
    _null_trans = string.maketrans("", "")

    if "\0" in s:
        return False
    
    if not s:  # Empty files are considered text
        return True

    # Get the non-text characters (maps a character to itself then
    # use the 'remove' option to get rid of the text characters.)
    t = s.translate(_null_trans, text_characters)

    # If more than 30% non-text characters, then
    # this is considered a binary file
    if len(t)/len(s) > 0.30:
        return False
    return True


def parse_command_line():
    usage = "Usage: %prog [options] <input configfile/map/datfile> <output configfile/map/datfile>"
    description = "Change orientation from icosahedral to a different symmetry. \
WARNING: currenty only rotation supported is from 532/I90 to one with \
vertex originally in xy plane, -x/y quadrant, to negative Z-axis. \
Input can be either a configuration (master/continue) file, a map or a orientation parameter(dat) file. \
The output name will be interpreted accordingly. \
Rules if a parameter file is given: the map and orientation files therein listed \
are manipulated consistently, and the user-given suffix is used to generate new names. \
If the inverse rotation is specified (--inverse) and the suffix \
is found in the original names, then the suffix is removed."
    parser = OptionParser(usage=usage, description=description)
    parser.add_option("-i","--inverse",action="store_true",dest="inverse_rot",help="apply inverse rotation, i.e. from 5-fold to icosahedral (default: no)", default=False)
    parser.add_option("-s","--suffix",dest="suffix",metavar="<string>",type="string",help="suffix to append to output reults (only when giving configuration files, default: _rot)",default="_rot")

    if len(sys.argv) == 1: 
        parser.print_help()
        sys.exit(-1)
	
    (options, args)=parser.parse_args()

    if len(args) < 2:
        print "\nPlease specify both input and output file. Type '%s --help' to see how to use it.\n"%os.path.basename(sys.argv[0])
        sys.exit(-1)
	    
    if not os.path.exists(args[0]):
        print "\nCannot find input file %s!\n"%args[0]
        sys.exit(-1)

    if os.path.exists(args[1]):
        print "\nWARNING:the file %s will be overwritten!\n"%args[1]

    return (options,args)


if __name__== "__main__":
    main()
