#!/usr/bin/env python
#title           :a3dm_mg_inspect.py
#description     :GUI to screen micrographs by visual evaluation
#author          :Giovanni Cardone (UCSD)
#date            :20150901
#version         :0.2
#notes           :experimental
#==============================================================================

import os, sys, shutil
import Tkinter
import tkMessageBox
has_ttk = True
try:
    from PIL import Image, ImageTk
except:
    import Image, ImageTk
try:
    import ttk
except:
    has_ttk = False

from pya3dm.mginspcfg import MgInspCfg, configfile

mgcnvsize = 960
pscnvsize = 512
ctfcnvsize = 256

class Screener:
    def __init__(self, imgsequence, mgDir, mgIgnoreDir, ctfDir, ctfIgnoreDir):
        self.imgseq =  imgsequence
        self.mgDir = mgDir
        self.ctfDir = ctfDir
        self.mgIgnoreDir = mgIgnoreDir
        self.ctfIgnoreDir = ctfIgnoreDir
        if not os.path.exists(self.mgDir): raise IOError("Directory %s not found!"%self.mgDir)
        if not os.path.exists(self.ctfDir): raise IOError("Directory %s not found!"%self.ctfDir)
        if not os.path.exists(self.mgIgnoreDir):  os.mkdir(self.mgIgnoreDir)
        if not os.path.exists(self.ctfIgnoreDir): os.mkdir(self.ctfIgnoreDir)

    def moveData(self):
        from glob import glob
        refimgs = self.imgseq.GetAllNames()
        nrefimgs = len(refimgs)

        mgNo = []
        ctfNo = []
        allmgavail = []
        mgInDir = []
        mgInIgnoreDir = []
        for x in ['mrc','pif']:
            for d in [self.mgDir]:
                mgInDir.extend(glob(os.path.join(d,"*.%s"%x)))
            for d in [self.mgIgnoreDir]:
                mgInIgnoreDir.extend(glob(os.path.join(d,"*.%s"%x)))
        allmgavail = mgInDir + mgInIgnoreDir
        allmgavail = [os.path.splitext(os.path.basename(x))[0] for x in allmgavail]

        allctfavail = []
        ctfInDir = []
        ctfInIgnoreDir = []
        for x in ['ctf']:
            for d in [self.ctfDir]:
                ctfInDir.extend(glob(os.path.join(d,"*.%s"%x)))
            for d in [self.ctfIgnoreDir]:
                ctfInIgnoreDir.extend(glob(os.path.join(d,"*.%s"%x)))
        allctfavail = ctfInDir + ctfInIgnoreDir
        allctfavail = [os.path.splitext(os.path.basename(x))[0] for x in allctfavail]
                
        for im in refimgs:
            # check if mg and ctf available
            mgavail = im in allmgavail
            ctfavail = im in allctfavail
            if not mgavail:  mgNo.append(im)
            if not ctfavail: ctfNo.append(im)
            if mgavail:
                if self.imgseq.isTagged(im):
#                    if any(im in s for s in mgInDir):
#                        indices = [i for i, s in enumerate(mgInDir) if im in s]
                    if any(im==os.path.splitext(os.path.basename(s))[0] for s in mgInDir):
                        indices = [i for i, s in enumerate(mgInDir) if im==os.path.splitext(os.path.basename(s))[0]]
                        if len(indices)>1:
                            raise ValueError("ERROR: ambiguity in determining micrograph name (%s)!"%im)
                        mgname = mgInDir[indices[0]]
                        shutil.move(mgname,self.mgIgnoreDir)
                else:
#                    if any(im in s for s in mgInIgnoreDir):
#                        indices = [i for i, s in enumerate(mgInIgnoreDir) if im in s]
                    if any(im==os.path.splitext(os.path.basename(s))[0] for s in mgInIgnoreDir):
                        indices = [i for i, s in enumerate(mgInIgnoreDir) if im==os.path.splitext(os.path.basename(s))[0]]
                        if len(indices)>1:
                            raise ValueError("ERROR: ambiguity in determining micrograph name (%s)!"%im)
                        mgname = mgInIgnoreDir[indices[0]]
                        shutil.move(mgname,self.mgDir)
            if ctfavail:
                if self.imgseq.isTagged(im):
#                    if any(im in s for s in ctfInDir):
#                        indices = [i for i, s in enumerate(ctfInDir) if im in s]
                    if any(im==os.path.splitext(os.path.basename(s))[0] for s in ctfInDir):
                        indices = [i for i, s in enumerate(ctfInDir) if im==os.path.splitext(os.path.basename(s))[0]]
                        if len(indices)>1:
                            raise ValueError("ERROR: ambiguity in determining ctf name (%s)!"%im)
                        ctfname = ctfInDir[indices[0]]
                        shutil.move(ctfname,self.ctfIgnoreDir)
                else:
#                    if any(im in s for s in ctfInIgnoreDir):
#                        indices = [i for i, s in enumerate(ctfInIgnoreDir) if im in s]
                    if any(im==os.path.splitext(os.path.basename(s))[0] for s in ctfInIgnoreDir):
                        indices = [i for i, s in enumerate(ctfInIgnoreDir) if im==os.path.splitext(os.path.basename(s))[0]]
                        if len(indices)>1:
                            raise ValueError("ERROR: ambiguity in determining ctf name (%s)!"%im)
                        ctfname = ctfInIgnoreDir[indices[0]]
                        shutil.move(ctfname,self.ctfDir)
                    
        return nrefimgs, len(allmgavail), mgNo, len(allctfavail), ctfNo                
        
class ImageSequence:
    def __init__(self,imgsdir,imgExt,mgSfx,psSfx,ctfSfx):
        self.tagPrefix = ".delete_"
        self.imgList= []
        self.tagList= []
        self.mgDict = {}
        self.psDict = {}
        self.ctfDict = {}
        self.defDict = {}
        self.imgsdir = imgsdir
        self.idx = -1
        listing = os.listdir(imgsdir)
        #and makes a dict of all images in the directory
        for item in listing:
            imgfile = os.path.basename(item)
            imgfullname, ext = os.path.splitext(imgfile)
            if imgExt in ext:
                imgname = ''
                if imgfullname.endswith(mgSfx):
                    imgname = imgfullname[:-len(mgSfx)]
                    if not imgname in self.mgDict.keys():
                        self.mgDict[imgname] = item
                elif imgfullname.endswith(psSfx):
                    imgname = imgfullname[:-len(psSfx)]
                    if not imgname in self.psDict.keys():
                        self.psDict[imgname] = item
                elif imgfullname.endswith(ctfSfx):
                    imgname = imgfullname[:-len(ctfSfx)]
                    if not imgname in self.ctfDict.keys():
                        self.ctfDict[imgname] = item
                if imgname and not imgname in self.imgList:
                    self.imgList.append(imgname)
                    tagfile = os.path.join(self.imgsdir,self.tagPrefix+imgname)
                    if os.path.exists(tagfile):
                        self.tagList.append(True)
                    else:
                        self.tagList.append(False)
                    deffile = os.path.join(self.imgsdir,imgname+'.ctf')
                    if os.path.exists(deffile):
                        vals = (file(deffile).readlines()[0]).split()
                        davg,dast,dcc = vals[:3]
                        if len(vals) == 4:
                            reslim = vals[3]
                            self.defDict[imgname] = ('%0.2f'%float(davg),'%0.2f'%float(dast),'%0.3f'%float(dcc),'%0.1f'%float(reslim)) 
                        else:
                            self.defDict[imgname] = ('%0.2f'%float(davg),'%0.2f'%float(dast),'%0.3f'%float(dcc), '-') 
                    else:
                        self.defDict[imgname] = ('-','-','-','-')
                
        #sorts images into order
        sorted_lists = sorted(zip(self.imgList, self.tagList), reverse=False, key=lambda x: x[0])
        self.imgList, self.tagList = [[x[i] for x in sorted_lists] for i in range(2)]
        if not self.imgList:
            print("ERROR: no images found!")
            sys.exit(-1)

    def isTagged(self, img = None):
        if img:
            idx = self.imgList.index(img)
        else:
            idx = self.idx
        return self.tagList[idx]

    def tagName(self, img = None):
        if not img: img = self.GetName()
        if self.isTagged(img): return
        if not img:
            img = self.imgList[self.idx]
        tagfile = os.path.join(self.imgsdir,self.tagPrefix+img)
        if not os.path.exists(tagfile):
            open(tagfile, 'a').close()
        idx = self.imgList.index(img)
        self.tagList[idx] = True

    def untagName(self, img = None):
        if not img: img = self.GetName()
        if not self.isTagged(img): return
        tagfile = os.path.join(self.imgsdir,self.tagPrefix+img)
        if os.path.exists(tagfile):
            os.remove(tagfile)
        idx = self.imgList.index(img)
        self.tagList[idx] = False

    def isLast(self):
        return self.idx+1==len(self.imgList)

    def isFirst(self):
        return (self.idx==0)

    def GetTotal(self, tag='all'):
        if tag == 'all':
            return len(self.imgList)
        elif tag == 'tagged':
            return self.tagList.count(True)
        elif tag == 'untagged':
            return self.tagList.count(False)
        return 0
    
    def GetIndex(self, tag='all'):
        if tag == 'all':
            return self.idx+1
        elif tag == 'tagged':
            return self.tagList[:self.idx].count(True)+1
        elif tag == 'untagged':
            return self.tagList[:self.idx].count(False)+1
        return 0

    def GetState(self):
        return self.idx
    
    def SetState(self,state):
        if self.imgList:
            self.idx=state

    def SetName(self,name):
        if self.imgList:
            self.idx=self.imgList.index(name)

    def GetName(self):
        if not self.imgList: return None
        return self.imgList[self.idx]
    
    def GetAllNames(self):
        return self.imgList
    
    def GetFirst(self):
        if not self.imgList: return None
        self.idx = 0
        return self.__get()
    
    def GetPrev(self):
        if not self.imgList: return None
        if self.idx > 0:
            self.idx = self.idx - 1
        return self.__get()

    def GetNext(self):
        if not self.imgList: return None
        if self.idx < (len(self.imgList)-1):
            self.idx += 1
        return self.__get()

    def GetCurr(self):
        if not self.imgList: return None
        return self.__get()

    def __get(self):
        nm = self.imgList[self.idx]
        mgname = psname = ctfname = None
        if nm in self.mgDict.keys():
            mgname = os.path.join(self.imgsdir,self.mgDict[nm])
        if nm in self.psDict.keys():
            psname = os.path.join(self.imgsdir,self.psDict[nm])
        if nm in self.ctfDict.keys():
            ctfname = os.path.join(self.imgsdir,self.ctfDict[nm])

        return (mgname, psname, ctfname)            

    def GetDefocus(self):
        img = self.GetName()
        return self.defDict[img] 
                
class GUI(Tkinter.Frame):

    def updateCanvas(self,mgname,psname,ctfname):
        global mgimg_tk, psimg_tk, ctfimg_tk
        if mgname:
            mgimg = Image.open(mgname)
            mgimg_tk = ImageTk.PhotoImage(mgimg)
            self.mgCanvas.create_image(mgcnvsize//2, mgcnvsize//2, image=mgimg_tk)
        else:
            self.mgCanvas.delete("all")
        self.mgCanvas.update()
        if psname:
            psimg = Image.open(psname)
            psimg_tk = ImageTk.PhotoImage(psimg)
            self.psCanvas.create_image(pscnvsize//2, pscnvsize//2, image=psimg_tk)
        else:
            self.psCanvas.delete("all")
        self.psCanvas.update()
        if ctfname:
            ctfimg = Image.open(ctfname)
            ctfimg_tk = ImageTk.PhotoImage(ctfimg)
            self.ctfCanvas.create_image(ctfcnvsize//2, ctfcnvsize//2, image=ctfimg_tk)
        else:
            self.ctfCanvas.delete("all")
        self.ctfCanvas.update()

    def isToShow(self):
        showmode = self.showmode.get()
        if showmode == 'all': return True
        if self.images.isTagged() and showmode == 'onlyIgnore':
            return True
        if not self.images.isTagged() and showmode == 'onlyKeep':
            return True
        return False
            
    def NextImage(self,event=None):
        currentstate = self.images.GetState()
        toupdate = False
        while not self.images.isLast():
            mgname, psname, ctfname = self.images.GetNext()
            if self.isToShow():
                toupdate =True
                break
        if toupdate:
            self.updateImageLabel()
            self.updateCanvas(mgname,psname,ctfname)
        else:
            self.images.SetState(currentstate)

    def PrevImage(self,event=None):
        currentstate = self.images.GetState()
        toupdate = False
        while not self.images.isFirst():
            mgname, psname, ctfname  = self.images.GetPrev()
            if self.isToShow():
                toupdate = True
                break
        if toupdate:
            self.updateImageLabel()
            self.updateCanvas(mgname,psname,ctfname)
        else:
            self.images.SetState(currentstate)
        
    def UpdateImage(self,event=None):
        self.ClosestImage()
        self.updateImageLabel()
    
    def ClosestImage(self):
        showmode = self.showmode.get()
        if showmode == 'all': return
        if self.isToShow(): return
        currentstate = self.images.GetState()
        self.NextImage()
        if self.images.GetState() == currentstate:
            self.PrevImage()
        if self.images.GetState() == currentstate:
            warnmsg = 'Undefined'
            if showmode == 'onlyIgnore':
                warnmsg = "No images to ignore"
            if showmode == 'onlyKeep':
                warnmsg = "No images to keep"
            tkMessageBox.showwarning(
            "Show", warnmsg)
            self.showmode.set("all")

    def KeepImage(self):
        self.images.untagName()
        self.imagenamelabel["bg"] = "green"
        if not self.isToShow(): self.ClosestImage()
                
    def IgnoreImage(self,event=None):
        self.images.tagName()
        self.imagenamelabel["bg"] = "red"
        if not self.isToShow(): self.ClosestImage()

    def ScreenData(self):
        mgDir = self.presets.mgDir
        mgIgnoreDir = self.presets.mgIgnoreDir
        ctfDir = self.presets.ctfDir
        ctfIgnoreDir = self.presets.ctfIgnoreDir
        try:
            screener = Screener(self.images, mgDir, mgIgnoreDir, ctfDir, ctfIgnoreDir)
            (nimgs, nmgavail, mgmissing, nctfavail, ctfmissing) = screener.moveData()
        except IOError as e:
            errmsg = str(e)
            tkMessageBox.showerror("Screen", errmsg)
        except ValueError as v:
            warnmsg = str(v)
            tkMessageBox.showerror("Screen", warnmsg)
        warnmsg = ""
        if len(mgmissing) > 0:
            warnmsg += "%d micrograph(s) were not found!\n"%len(mgmissing)
        if len(ctfmissing) > 0:
            warnmsg += "%d CTF estimate(s) were not found!\n"%len(ctfmissing)
        if warnmsg:
            tkMessageBox.showwarning("Screen", warnmsg)
        else:
            tkMessageBox.showinfo("Screen", "Data were successfully moved!")

    def updateImageLabel(self):
        currname = self.images.GetName()
        if has_ttk: self.mgchoicevalue.set(currname)
        tagtype = self.tagtype[self.showmode.get()]
        idx = self.images.GetIndex(tagtype)
        N = self.images.GetTotal(tagtype)
        if self.images.isTagged():
            self.imagenamelabel["bg"] = "red"
        else:
            self.imagenamelabel["bg"] = "green"
        self.imagename.set("%s (%d/%d)"%(currname,idx,N))

        defavg, defast, defcc, defreslim = self.images.GetDefocus()
        self.defocus.set("Defocus (avg / %%astig / fom / reslim): %s um / %s%% / %s / %s A"%(defavg, defast, defcc,defreslim))
        
    def chooseMg(self, event):
        mg = self.mgchoicevalue.get()
        self.showmode.set('all')
        self.images.SetName(mg)
        mgname, psname, ctfname  = self.images.GetCurr()
        self.updateImageLabel()
        self.updateCanvas(mgname,psname,ctfname)
        
    def CreateCanvas(self):
        global mgimg_tk, psimg_tk, ctfimg_tk

        self.topframe=Tkinter.Frame(self)
        self.topframe.pack({"side":"top","fill":"x"})

        self.leftframe=Tkinter.Frame(self.topframe)
        self.leftframe.pack({"side":"left","fill":"x"})

        mgname, psname, ctfname = self.images.GetFirst()

        self.mgCanvas = Tkinter.Canvas(self.leftframe, width=mgcnvsize, height=mgcnvsize)
        self.mgCanvas.pack({"side":"top"})

        self.rightframe=Tkinter.Frame(self.topframe)
        self.rightframe.pack({"side":"right","fill":"x"})

        self.imagenamelabel=Tkinter.Label(self.rightframe)
        self.imagename = Tkinter.StringVar()
        self.imagenamelabel["textvariable"]=self.imagename
        self.imagenamelabel.pack({"side":"top"})

        self.defocuslabel=Tkinter.Label(self.rightframe)
        self.defocus = Tkinter.StringVar()
        self.defocuslabel["textvariable"]=self.defocus
        self.defocuslabel.pack({"side":"top"})

        self.psCanvas = Tkinter.Canvas(self.rightframe, width=pscnvsize, height=pscnvsize)
        self.psCanvas.pack({"side":"top"})

        self.ctfCanvas = Tkinter.Canvas(self.rightframe, width=ctfcnvsize, height=ctfcnvsize)
        self.ctfCanvas.pack({"side":"top"})

        self.updateCanvas(mgname,psname,ctfname)

        
    def CreateButtons(self):

        self.bottomframe=Tkinter.Frame(self.rightframe)
        self.bottomframe.pack({"side":"bottom"})
        

        self.prev=Tkinter.Button(self.bottomframe)
        self.prev["text"]="Prev"
        self.prev["fg"]="blue"
        self.prev["command"]=self.PrevImage
        self.prev.pack({"side":"left"})

        self.next=Tkinter.Button(self.bottomframe)
        self.next["text"]="Next"
        self.next["fg"]="blue"
        self.next["command"]=self.NextImage
        self.next.pack({"side":"left"})

        self.ignore=Tkinter.Button(self.bottomframe)
        self.ignore["text"]="Ignore"
        self.ignore["fg"]="red"
        self.ignore["command"]=self.IgnoreImage
        self.ignore.pack({"side":"left"})

        self.keep=Tkinter.Button(self.bottomframe)
        self.keep["text"]="Keep"
        self.keep["fg"]="green"
        self.keep["command"]=self.KeepImage
        self.keep.pack({"side":"left"})

        self.screen=Tkinter.Button(self.bottomframe)
        self.screen["text"]="Screen"
        self.screen["fg"]="black"
        self.screen["command"]=self.ScreenData
        self.screen.pack({"side":"left"})
        
        self.showmode = Tkinter.StringVar()
        self.showmode.set("all")
        modes = [ ("all","all"), ("to ignore","onlyIgnore"), ("to keep","onlyKeep")]

        self.showframe = Tkinter.Frame(self.bottomframe, bd=2, relief=Tkinter.SUNKEN)
        self.showlabel=Tkinter.Label(self.showframe)
        self.showlabel["text"]="Show"
        self.showlabel.pack({"side":"left"})
        
        for txt, val in modes:
            Tkinter.Radiobutton(self.showframe, command=self.UpdateImage, text=txt, padx = 5, variable=self.showmode, value=val).pack(anchor=Tkinter.W)
        self.showframe.pack({"expand":"1", "fill":"x", "pady":"10", "padx":"5"})

        if has_ttk:
            self.listframe = Tkinter.Frame(self.leftframe)
            self.listframe.pack({"side":"bottom"})
            self.mgchoicevalue = Tkinter.StringVar()
            self.choicelabel = Tkinter.Label(self.listframe)
            self.choicelabel["text"]="Jump to: "
            self.choicelabel.pack({"side":"left"})
            cbwidth = min(max([len(x) for x in self.images.GetAllNames()]),mgcnvsize)
            self.choicelist  = ttk.Combobox(self.listframe, textvariable=self.mgchoicevalue, values=self.images.GetAllNames(), state='readonly', height=10, width=cbwidth)
            self.choicelist.pack({"side":"left"}) #, {"fill":"yes"})
            self.choicelist.current(0)
#            self.choicelist.grid(column=0, row=0)
            self.choicelist.bind('<<ComboboxSelected>>', self.chooseMg)


    def __init__(self, master=None,start_dir=None):
        Tkinter.Frame.__init__(self, master)
        self.master = master
        self.master.bind('<Left>', self.PrevImage)
        self.master.bind('<Right>', self.NextImage)
        self.master.bind('<Delete>', self.IgnoreImage)
        self.pack()
        self.imgdir = start_dir
        self.presets = MgInspCfg(os.path.join(start_dir,configfile))
        self.images = ImageSequence(start_dir,self.presets.imgExt,self.presets.mgSfx,self.presets.psSfx,self.presets.ctfSfx)
        self.tagtype = {'all':'all', 'onlyIgnore':'tagged', 'onlyKeep':'untagged'}
        self.CreateCanvas()
        self.CreateButtons()
        self.updateImageLabel()

		

def main(start_dir):
    global mgcnvsize, pscnvsize, ctfcnvsize
    root = Tkinter.Tk()
    screen_width = root.winfo_screenwidth()
    screen_height = root.winfo_screenheight()
    xfactor = 1.
    while int(xfactor*mgcnvsize>(screen_height*0.9)):
        xfactor -= 0.1
    if xfactor < 1.:
        mgcnvsize = int(mgcnvsize * xfactor)
        pscnvsize  = int(pscnvsize * xfactor)
        ctfcnvsize = int(ctfcnvsize * xfactor)
           
#    root.geometry("300x280+300+300")
    root.title("AUTO3DEM - inspect micrographs")
    app = GUI(root,start_dir)
    root.mainloop()  


if __name__ == '__main__':

    if len(sys.argv)<2:
        start_dir = os.getcwd()
    else:
        start_dir = sys.argv[1]
        if not os.path.exists(start_dir):
            print("Can not find input directory %s!"%start_dir)
            sys.exit(-1)
    main(start_dir)  
