#!/usr/bin/env python
#title           :a3dm_mg_inspect_prepare.py
#description     :Generate image files for screening purposes
#author          :Giovanni Cardone (UCSD)
#date            :20150713
#version         :0.2
#notes           :experimental
#==============================================================================
import os, sys, shutil, re, string, subprocess
from glob import glob
from optparse import OptionParser

from pya3dm.mginspcfg import MgInspCfg, configfile

def get_map_size(fname):
    nx = ny = nz = 0
    proc = subprocess.Popen('bhead -verb 4 -info '+fname, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    output = stdout.splitlines()
    errors = str(stderr)
    if proc.returncode > 0 and not 'Warning: Swapping header byte order for 4-byte types' in errors:
        print('FILE '+fname+':')
        raise RuntimeError(errors)
    for l in output:
        if (l.strip()).startswith('Dimensions'):
            xl = string.split(l)
            x = int(xl[1])
            ny = int(xl[2])
            nz = int(xl[3])
            break

    return nx, ny, nz

def get_map_stats(fname):
    avg = std = mmin = mmax = 0
    proc = subprocess.Popen('bhead -v 4 -info -recalc '+fname, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    output = stdout.splitlines()
    errors = str(stderr)
    if proc.returncode > 0 and not 'Warning: Swapping header byte order for 4-byte types' in errors:
       print('FILE '+fname+':')
       raise RuntimeError(errors)
    for l in output:
       if re.search('Min,', l):
          xl = string.split(l)
          mmin = float(xl[4])
          mmax = float(xl[5])
          avg = float(xl[6])
          std = float(xl[7])
          break

    return mmin, mmax, avg, std

def pow_screen_images(powpowfiles,powmrcfiles):
    ''' 
Assuming both ctffind3 and ctffind4 have been run on the micrographs,
find duplicated info and keep images from ctffind3, since so far they
are more useful for visual evaluation
    ''' 
    if not powpowfiles:
        return powmrcfiles
    if not powmrcfiles:
        return powpowfiles

    powfiles = powpowfiles[:]
    candidates = [ os.path.basename(x[:-8]) for x in powmrcfiles ]
    references = [ os.path.basename(x[:-4]) for x in powpowfiles ]
    for i,x in enumerate(candidates):
        if not x in references:
            powfiles.append(powmrcfiles[i])

    return powfiles

def main():
    (options,args) =  parse_command_line()

    outdir = args[0]
    
    mgdir = options.mg_dir
    ctfdir = options.ctf_dir
    
# Generate a config file for screening during visual inspection
    mgcfg = MgInspCfg()
    imgExt = mgcfg.imgExt
    mgSfx = mgcfg.mgSfx
    psSfx = mgcfg.psSfx
    ctfSfx = mgcfg.ctfSfx
        
    mgcfg.mgDir = os.path.abspath(mgdir)
    mgcfg.ctfDir = os.path.abspath(ctfdir)
    mgcfg.mgIgnoreDir = os.path.abspath(mgdir)+'_ignore'
    mgcfg.ctfIgnoreDir = os.path.abspath(ctfdir)+'_ignore'

    mgcfg.write(os.path.join(outdir,configfile))    
    
# Generate input data and check consistency
    mgfiles = glob(os.path.join(mgdir,"*.mrc"))+glob(os.path.join(mgdir,"*.pif"))
    nmg = len(mgfiles)
    
    print("Found %d micrographs"%nmg)

    powpowfiles = glob(os.path.join(ctfdir,"*.pow"))
    powmrcfiles = glob(os.path.join(ctfdir,"*_pow.mrc"))
    powfiles = pow_screen_images(powpowfiles,powmrcfiles)
    npow = len(powfiles)
    ctffiles = glob(os.path.join(ctfdir,"*.ctf"))
    nctf = len(ctffiles)
    
    nxctf = max(npow,nctf)
    print("Found %d ctf images"%nxctf)
    if nmg != nctf:
        print("WARNING: number of micrographs (%d) is different from number of ctf images (%d)"%(nmg,nctf))
        #sys.exit(-1)    
    
# Generate pngs from micrographs

    print("\nPreparing micrograph images\n")
    
    # first determine binning
    if mgfiles:
        bin = 2
        target_size = 960
        nx,ny,nz = get_map_size(mgfiles[0])
        while (nx/bin > target_size) or (ny/bin > target_size) or (nz/bin > target_size):
            bin+=1
        print(" *** Binning factor: %d ***"%bin)

    for i,p in enumerate(mgfiles):
        print("%d/%d - %s"%(i+1,nmg,p))
        pngf = os.path.join(outdir,os.path.splitext(os.path.basename(p))[0]+mgSfx+'.'+imgExt)
        if not os.path.exists(pngf):
            tmp = os.path.join(outdir,"tmp.mrc")
            os.system("bint -bin %d %s %s"%(bin,p,tmp))
            mmin, mmax, avg, std = get_map_stats(tmp)
            if mmin < 0.:
                os.system("bimg -data int -minmax 0,32767 %s %s"%(tmp,tmp)) 
            os.system("bimg -data ushort %s %s"%(tmp,pngf))
            os.system("bimg -data byte -minmax 0,255 %s %s"%(pngf,pngf))
            os.remove(tmp)
        else: print "... thumbnail image already available. Skipping"


# Generate pngs from ctf fit

    print("\nPreparing CTF images\n")

    for i,p in enumerate(powfiles):
        print("%d/%d - %s"%(i+1,npow,p))
        if '_pow.mrc' in p:
            outpfx = os.path.splitext(os.path.basename(p))[0][:-4]
        else: 
            outpfx = os.path.splitext(os.path.basename(p))[0]
        pngf = os.path.join(outdir,outpfx+ctfSfx+'.'+imgExt)
        if not os.path.exists(pngf):
            os.system("bimg -data ushort %s:mrc %s"%(p,pngf))
            os.system("bimg -data byte -minmax 0,255 %s %s"%(pngf,pngf))
        else: print "... png image already available. Skipping"

    print("\nPreparing CTF info\n")

    for i,p in enumerate(ctffiles):
        print("%d/%d - %s"%(i+1,nctf,p))
        ctff = os.path.join(outdir,os.path.splitext(os.path.basename(p))[0]+'.ctf')
        if os.path.exists(ctff):
            print "... ctf info already available. Skipping"
            continue
        flines = file(p).readlines()
        cc = None
        reslim = None
        for x in flines:
            if "Final Values" in x:
                fvals = x.split()
                df1 = float(fvals[0])
                df2 = float(fvals[1])
                cc = float(fvals[3])
            if "Thon" in x:
                fvals =x.split()
                reslim = float(fvals[8])
                break
        if not cc:
            print("CTF info from %s not available. Skipping..."%p)
        else:
            dfavg = 0.5*(df1+df2)/10000.
            dfast = (max(df1,df2)/min(df1,df2)-1.)*100
            if reslim:
                outline = ["%f %f %f %f\n"%(dfavg,dfast,cc,reslim)]
            else:
                outline = ["%f %f %f\n"%(dfavg,dfast,cc)]
            file(ctff,'w').writelines(outline)       
            
# Generate power spectrum from micrographs

    print("\nPreparing spectrum images\n")

    for i,p in enumerate(mgfiles):
        print("%d/%d - %s"%(i+1,nmg,p))
        pngf = os.path.join(outdir,os.path.splitext(os.path.basename(p))[0]+psSfx+'.'+imgExt)
        if not os.path.exists(pngf):
            tmpf = os.path.join(outdir,os.path.splitext(os.path.basename(p))[0]+'_ps.mrc')
            os.system("bfft -data float -power -tile 1024,1024,1 %s %s"%(p,tmpf))
            os.system("bint -bin 2 %s %s"%(tmpf,tmpf))
            mmin, mmax, avg, std = get_map_stats(tmpf)
            os.system("beditimg -sphere 256,256,0,12 -fill %f %s %s"%(avg,tmpf,tmpf))
            mmin, mmax, avg, std = get_map_stats(tmpf)
            os.system("bimg -trunc %f,%f %s %s"%(avg-5.*std,avg+5.*std,tmpf,tmpf))
            os.system("bimg -data ushort %s %s"%(tmpf,pngf))
            os.system("bimg -data byte -minmax 0,255 %s %s"%(pngf,pngf))
#            os.system("bfft -data float -power -tile 1024,1024,1 %s %s"%(p,tmpf))
#            os.system("bint -bin 2 %s %s"%(tmpf,tmpf))
#            os.system("bimg -data ushort %s %s"%(tmpf,pngf))
#            mmin, mmax, avg, std = get_map_stats(tmpf)
#            os.system("beditimg -sphere 256,256,0,12 -fill %f %s %s"%(avg,tmpf,pngf))
#            mmin, mmax, avg, std = get_map_stats(tmpf)
#            os.system("bimg -trunc %f,%f %s %s"%(avg-5.*std,avg+5.*std,pngf,pngf))
#            os.system("bimg -data byte -minmax 0,255 %s %s"%(pngf,pngf))
            os.remove(tmpf)
        else: print "... png image already available. Skipping"

    print "Done!"



def parse_command_line():
    usage = "Usage: %prog [options] <output directory>"
    description = "Prepare thumbnail images in a user-defined directory to use for evaluation.\nSpecifically, the images are: micrograph, its spectrum and the agreement between theoretical and experimental ctf."
    parser = OptionParser(usage=usage, description=description)
    parser.add_option("-m","--micrographs",dest="mg_dir",metavar="<mg dir>",type="string",help="Directory containing only the micrographs in mrc or pif format(default: ./mg)",default="./mg")
    parser.add_option("-c","--ctfs",dest="ctf_dir",metavar="<ctf dir>",type="string",help="Directory containing the CTF parameters estimated by CTFFIND3 (default: ./ctf)",default="./ctf")
    

    if len(sys.argv)<2: 
        parser.print_help()
        print
        sys.exit(-1)
	
    (options, args)=parser.parse_args()

    if (not args):
        print "\nERROR: output directory needed!\n"
        sys.exit(-1)

    if not os.path.exists(args[0]):
        print "Creating directory %s"%args[0]
        os.mkdir(args[0])

    return (options,args)


if __name__== "__main__":
    main()
