#!/usr/bin/perl

use strict;
use warnings;
use FindBin qw($Bin);
use lib "$Bin/modules";
use lib "$Bin/";

# The following block of code determines whether or not the program
# Xmgrace (gracebat is simply the batch interface to Xmgrace) is in
# the users path. The Grace .agr file will be generated regardless of
# the result of the test.

my $gracebat;
my @path = split(/:/, $ENV{PATH});
foreach my $path (@path) {
    if(-e "$path/gracebat") {
	$gracebat = "$path/gracebat";
	last;
    }
}

use File::Copy;
use Cwd;
use make_program_input;
use init_params;
use init_cpus;
use init_launch_cmd;
use write_params;
use run_mpi_prog;
use incr;
use validate;
use psf_parse;
use write_header;
use update_res;
use get_info;
use select;
#use get_ann_lo_hi;
use sendmail;
use sanity_checks;
use mode;
use grace_utils;
use version;
use rad_profile_parse;

use constant PI    => 4 * atan2(1, 1);
sub max ($$) { $_[$_[0] < $_[1]] };
sub min ($$) { $_[$_[0] > $_[1]] };

my $time_start = time();
my $startdir = `pwd`;
my $mode = mode();
my $version = get_version();
my $version_string = "------ AUTO3DEM version $version ($mode) ------";
chomp($startdir);

# Print usage info if no arguments provided
if(@ARGV == 0 || $ARGV[0] =~ /^-help/i) {
print << "EOF";
$version_string

NAME
  auto3dem - automated 3D image reconstruction program

EXAMPLE
  auto3dem -[ncpu|np] 12 -input input_file -nodefile mynodelist

DESCRIPTION

  All input specified using the syntax -key value. -ncpu (MPI version
  only) and -input required; -nodefile optional. 'n' = integer value,
  's' = string

  -input s     file containing auto3dem input parameters
  -ncpu n      number of CPUs
  -np   n      alternative way of specifying number of CPUs
  -nodefile s  file containing list of compute nodes

NOTES
  Nodefile specification depends on your computing environment and in
  many cases it can be omitted. For systems running the PBS scheduler,
  the nodefile is automatically obtained from the environment assuming
  that "#PBS -V" was used in the batch script. Keys are case
  insensitive (e.g. -ncpu, -Ncpu, -NCPU all OK), but string values are
  case sensitive.
EOF
exit(0);
}

# Process command line args so that we can split on ' -'
my $command_line = join ' ', @ARGV;
$command_line =~ s/^\s+//;   # Get rid of leading whitespace
$command_line =~ s/\s+$//;   # Get rid of trailing whitespace
$command_line =~ s/=/ /g;    # Change '=' to ' '
$command_line =~ s/^-//;     # Get rid of leading '-'
@ARGV = split(/\s+-/, $command_line);

my %cmdline_arg;
foreach my $arg (@ARGV) {
    my($key, $value) = split(/\s+/, $arg);
    $key =~ tr/A-Z/a-z/;
    $cmdline_arg{$key} = $value;
}

my ($ncpu, $nodefile, $param_file);

if(defined $cmdline_arg{ncpu}) {
    $ncpu = $cmdline_arg{ncpu};
} elsif (defined $cmdline_arg{np}) {
    $ncpu = $cmdline_arg{np};
}

if(defined $cmdline_arg{nodefile}) {
    $nodefile = $cmdline_arg{nodefile};
    unless($nodefile =~ /^\//) {
	$nodefile = "$startdir/$nodefile";
    }
} elsif (defined $ENV{PBS_NODEFILE}) {
    $nodefile = $ENV{PBS_NODEFILE};
} elsif (defined $ENV{SGE_NODEFILE}) {
    $nodefile = $ENV{SGE_NODEFILE};
}

if(defined $cmdline_arg{input}) {
    $param_file = $cmdline_arg{input};
}

if($mode eq 'serial') {
    $ncpu = 1;
}
validate_commandline_args($ncpu, $param_file, $mode);

# Initialize PO2R, P3DR, PCUT, PSF, PPFT, PCTFR, and control hashes
my (%auto_param, %po2r_param, %p3dr_param,
    %pcut_param, %psf_param, %ppft_param, %pctfr_param);
my %ptlefiles;
init_params($param_file, \%auto_param, \%po2r_param,
	    \%p3dr_param, \%pcut_param, \%psf_param,
	    \%ppft_param, \%pctfr_param, \%ptlefiles);
# Open outfile and write header information
if($auto_param{restart}) {
    open(OUT, "+>>$auto_param{outfile}_log");
} else {
    open(OUT, ">$auto_param{outfile}_log");
}

print OUT "$version_string\n\n";
write_header($ncpu, $nodefile, \%auto_param, 
	     \%p3dr_param, \%po2r_param, 
	     \%pcut_param, \%psf_param, \%ppft_param);

# Open summary output file and fill array of estimated resolutions if
# this is a restart calculation

my @estres = ();

my $summary_file = "$auto_param{outfile}_summary";
if($auto_param{restart} && -e $summary_file) {
    open(SUM, "+>>$summary_file");
    seek(SUM, 0, 0); # Move to start of file
    while (<SUM>) {
	s/^\s+//;
	my ($iter, $mode, $estres) = split(/\s+/);
	if(defined $iter && defined $mode && defined $estres && $estres !~ /-/ &&
	   $iter =~ /\d/ && ($mode =~ /s/ || $mode =~ /r/)) {
	    $estres =~ s/\*//; # Get rid of asterisk
	    unshift(@estres, $estres);
	}
    }
} else {
    # Create backup to avoid writing over old summary files
    if(-e $summary_file) {
	my $backup = "${summary_file}_backup";
	move $summary_file, $backup;
    }
    open(SUM, ">$summary_file");
    print SUM "$version_string\n\n";
    print SUM "itr:          iteration number\n";
    print SUM "              0 = starting map constructed from images\n";
    print SUM "mode:         s(n) = search mode (PPFT)  w/ n=binfactor\n";
    print SUM "              r(l) = refine mode (PO2R)  local search\n";
    print SUM "              r(g) = refine mode (PO2R)  global search\n";
    print SUM "              r(m) = refine mode (PO2R)  only magnification\n";
    print SUM "              r(c) = refine mode (PCTFR) only CTF\n";
    print SUM "estres:       estimated resolution of model\n";
    print SUM "              * = FSC never < $auto_param{fsc_cutoff}\n";
    print SUM "delta:        spacing between solutions tested (degrees/microns)\n";
    print SUM "map undamp:   resolution to which map is computed\n";
    print SUM "map damp:     resolution at which map density damped to zero\n";
    print SUM "time:         wallclock time for iteration\n";
    print SUM "cpu:          number of CPUs (MPI processes)\n";
    print SUM "nptles:       number ptles used to construct map\n";
    print SUM "ntot:         total number ptles in parameter files\n";
    print SUM "nmg:          number of micrographs\n";
    print SUM "defocus       defocus range (microns)\n";
    print SUM "\n";
    print SUM "All resolutions expressed in Angstroms, times in seconds\n";
    print SUM "\n";

    print SUM "                 delta   map    map\n";
    print SUM "itr mode estres  angle undamp  damp   time cpu nptles  ntot  nmg  defocus\n";
    print SUM "--- ---- ------ ------ ------ ------ ----- --- ------ ------ --- ---------\n";
}

# Change to the run directory where particle (and optionally map data)
# reside and where we want the results files (new particle
# orientations and maps) to be written

chdir($auto_param{rundir});
my $rundir = `pwd`;
chomp($rundir);

# Make sure that all of the particle parameter files are readable and
# properly formed

my $exit_flag = 0;

foreach my $subset (@{$ptlefiles{subsets}}) {
  foreach my $file (@{$ptlefiles{$subset}}) {
    unless(-r $file) {
	print "\n--- Errors for file $file --- \n";
	print "Particle parameter file is missing or not readable\n";
	$exit_flag = 1;
	next;
    }
    my ($warnings_ref, $errors_ref) = 
	param_file_check($file, $auto_param{noctf});
    if(@{$warnings_ref}) {
	print "\n--- Warnings for file $file ---\n";
	foreach my $warning (@{$warnings_ref}) {
	    print "$warning\n";
	}
    }
    if(@{$errors_ref}) {
	print "\n--- Errors for file $file --- \n";
	foreach my $error (@{$errors_ref}) {
	    print "$error\n";
	}
	$exit_flag = 1;
    }
    if($auto_param{gold_standard}==1) {
      my ($np, $nm, $dm, $dx, @stacknames) = get_paramfile_info($file);
      if($np==0) {
        print "ERROR: file $file does not contain particle parameters. Gold standard procedure require them.\n";
        $exit_flag = 1;
      }
    }
  }
}

# Exit if any errors were encountered
if($exit_flag) {
    exit(0);
}

# Create a directory where intermediate files can be moved
my $intfiles_dir = "$auto_param{outfile}_INTFILES";
mkdir $intfiles_dir;

my $firstptlefile = @{$ptlefiles{all}}[0];
# Get pixel_size from first ptlefile and update PSF pixel_size
my $pixel_size = get_pixel_size($firstptlefile);
$psf_param{pixel_size} = $pixel_size;

# Extract boxrad (packRadius in PIF standard) from image file
# Make undefined if value equal to zero
my $pif_name = get_img_name($firstptlefile);
my $boxrad_from_pif = get_boxrad($pif_name);
if($boxrad_from_pif == 0) {
    undef $boxrad_from_pif;
}

# The following code is used to allow for the specification of the box
# radius in either the PIF file or the auto3dem parameter
# file. Necessary because some PIF files are missing the box radius

# Case 1 - box radius not set in either location
unless(defined $auto_param{boxrad} || defined $boxrad_from_pif) {
    print OUT "Box radius not specified in PIF file or parameter file\n";
    print OUT "Either fix $pif_name or manually specify box radius\n";
    die "Terminating execution\n";
}

# Case 2 - box radius specified in both locations
if(defined $auto_param{boxrad} && defined $boxrad_from_pif) {
    print OUT "Box radius specified in both PIF file and parameter file\n";
    if($auto_param{boxrad} == $boxrad_from_pif) {
	print OUT "Box radius ($boxrad_from_pif) specified consistently\n\n";
    } else {
	print OUT "\nWarning boxrad from PIF file ($boxrad_from_pif) ";
	print OUT "not equal to boxrad from parameter file ";
	print OUT "($auto_param{boxrad})\n";
	print OUT "Value in parameter file overides PIF file\n\n";
    }
}

# Case 3 - box radius specified only in PIF file
if(defined $boxrad_from_pif && !defined $auto_param{boxrad}) {
    print OUT "Box radius from PIF file = $boxrad_from_pif\n\n";
    $auto_param{boxrad} = $boxrad_from_pif;
}

# Case 4 - box radius specified only in parameter file
if(!defined $boxrad_from_pif && defined $auto_param{boxrad}) {
    print OUT "Box radius from parameter file = $auto_param{boxrad}\n\n";
}

print OUT "Pixel size from $firstptlefile = $pixel_size\n\n";
print OUT "\n";

# Use boxrad and partrad to set program parameters

unless(defined $auto_param{partrad}) {
    $auto_param{partrad} = $auto_param{boxrad};
}

unless(defined $auto_param{box_center_offset}) {
    $auto_param{box_center_offset} = $auto_param{boxrad}/2;
}

unless(defined $pcut_param{in_rad}) {
    $pcut_param{in_rad} = sprintf "%0.0f", $auto_param{boxrad}/3.0;
}

unless(defined $pcut_param{out_rad}) {
    $pcut_param{out_rad} = sprintf "%0.0f", $auto_param{boxrad} - 10.0;
}

unless(defined $ppft_param{annulus_low}) {
    $ppft_param{annulus_low} = sprintf "%0.0f", $auto_param{boxrad}/3.0;
}

unless(defined $ppft_param{annulus_high}) {
    $ppft_param{annulus_high} = sprintf "%0.0f", $auto_param{boxrad} - 10.0;
}

unless(defined $ppft_param{pftrad_hi}) {
    $ppft_param{pftrad_hi} = $auto_param{boxrad};
}

unless(defined $ppft_param{resolution_low}) {
    $ppft_param{resolution_low} = $auto_param{partrad}*2.0*$pixel_size/5.0;
}

unless(defined $po2r_param{res_min}) {
    $po2r_param{res_min} = $auto_param{partrad}*2.0*$pixel_size/5.0;
}

unless(defined $pctfr_param{res_min}) {
    $pctfr_param{res_min} = $auto_param{partrad}*2.0*$pixel_size/5.0;
}

# Make sure that radii used by PPFT and PCUT are valid
my ($warnings_ref, $errors_ref) = 
    radii_check(\%auto_param, \%pcut_param, \%ppft_param);

if(@{$warnings_ref}) {
    print "\n--- Warnings for specification of PCUT and/or PPFT radii ---\n";
    foreach my $warning (@{$warnings_ref}) {
	print "$warning\n";
    }
}
if(@{$errors_ref}) {
    print "\n--- Errors for specification of PCUT and/or PPFT radii --- \n";
    foreach my $error (@{$errors_ref}) {
	print "$error\n";
    }
    $exit_flag = 1;
}

# Exit if any errors were encountered
if($exit_flag) {
    exit(0);
}

my $evod_parallel = $auto_param{evod_parallel};
if($evod_parallel == 1 && $ncpu == 1) {
    print OUT "\nINFO: parameter 'auto evod_parallel' forced to 0 (ncpu=1)\n\n";
    $evod_parallel = 0;
}

my ($po2r_cpu, $p3dr_cpu, $pcut_cpu, $psf_cpu, $ppft_cpu, $pctfr_cpu) =
   init_cpus($ncpu, $evod_parallel, \%po2r_param, \%p3dr_param, \%pcut_param, 
                      \%psf_param, \%ppft_param, \%pctfr_param);

# Set the command used to launch the single jobs.
my ($hostfiles, $launch_cmd) = init_launch_cmd(mode(), $p3dr_cpu, $rundir, $nodefile, $evod_parallel, $p3dr_param{cpu_sparse}); 
my @hostfiles = @$hostfiles;

if (! @hostfiles) {
  print OUT "Base command used for launching jobs\n";
  print OUT "$$launch_cmd{all}\n\n";
} else {
  print OUT "Base command used for launching jobs with all particles:\n";
  print OUT "$$launch_cmd{all}\n";
  print OUT "Base command used for launching jobs with odd particles:\n";
  print OUT "$$launch_cmd{odd}\n";
  print OUT "Base command used for launching jobs with even particles:\n";
  print OUT "$$launch_cmd{even}\n\n";
  print OUT "Base command used for launching reconstruction jobs with all particles:\n";
  print OUT "$$launch_cmd{p3drall}\n";
  print OUT "Base command used for launching reconstruction jobs with odd particles:\n";
  print OUT "$$launch_cmd{p3drodd}\n";
  print OUT "Base command used for launching reconstruction jobs with even particles:\n";
  print OUT "$$launch_cmd{p3dreven}\n\n";
}

# Save the originally specified number of iterations
my $niter_original = $auto_param{niter};

# Reduce to one iteration if only alignment or reconstruction requested
if($auto_param{generate_map} eq 'only') {
    $auto_param{niter} = 0;
} elsif($auto_param{generate_map} eq 'no') {
    $auto_param{niter} = 1;
}

# Determine the starting and ending iteration numbers
my ($first_iter, $last_iter);
if($auto_param{have_map} && $auto_param{generate_map} ne 'only') {
    $last_iter = $auto_param{iter_start} + $auto_param{niter} - 1;
} else {
    $auto_param{iter_start}--;
    $last_iter = $auto_param{iter_start} + $auto_param{niter};
}
$first_iter = $auto_param{iter_start};

# Set things up for flattening map
my %flatten_param;

# Set things up for hollowing map
my %hollow_param;
if($auto_param{hollow_map}) {
    $hollow_param{in_rad}        = $auto_param{hollow_in_rad};
    $hollow_param{out_rad}       = $auto_param{hollow_out_rad};
    $hollow_param{cut_step}      = $auto_param{hollow_cut_step}; 
    $hollow_param{cut_weight}    = $auto_param{hollow_cut_weight}; 
}

# Set the restart flag to true
$auto_param{restart} = 1;

# Maintain list of map files
my @mapfiles = ();
my @eomapfiles = ();
my $map_iter;
my $map_even_iter;
my $map_odd_iter;

my $capsid_score = 0.0;
# Initialize values from previous iteration with dummy values
my $prev_capsid_score = 0.0;
my $prev_po2r_angle = 100.0;
my $prev_ppft_angle = 100.0;

my $map_suffix = '';
if($auto_param{generate_map} eq 'only' && $auto_param{map_suffix} ne 'none') {
    $map_suffix = $auto_param{map_suffix};
}

# set flag for focused refinement, which needs generation of two soft-edge masks
my $target_refine = 0;
if($auto_param{mode} eq 'refine' && $po2r_param{mask_map} ne 'none') {
    unless(($auto_param{refine_ctf} && $auto_param{niter} == 1) || $auto_param{generate_map} eq 'only') {
        if($po2r_param{mode} eq 'local' || $po2r_param{mode} eq 'ticos_equiv') {
            $target_refine = 1;
        } else {
 	        print OUT "\nINFO: mask_map $po2r_param{mask_map} ignored because only used in local and ticos_equiv mode\n\n";
        }
    }
}    

my $pideven;
my $pidodd;
my $target_refine_iter = $target_refine;

# LOOP OVER ITERATIONS
for (my $iter=$first_iter; $iter<=$last_iter; $iter++) { 
    my $time_iter_start = time();
    my $nptles = 0;
    $map_iter = "$auto_param{outfile}_iter_${iter}${map_suffix}.pif";
    $map_even_iter = "$auto_param{outfile}_even_iter_${iter}${map_suffix}.pif";
    $map_odd_iter  = "$auto_param{outfile}_odd_iter_${iter}${map_suffix}.pif";

    print OUT "*********** Iteration $iter ***********\n\n";

    my $input_file;
    my $output_file;
    
    $auto_param{iter_start}++;

    # Summary information
    my $fscthresh_flag = " ";
    my $outmode = $auto_param{mode};
    my $outpo2rmode = $po2r_param{mode};
    $target_refine_iter = $target_refine;
	if($auto_param{refine_ctf} && $auto_param{mode} eq 'refine' && $pctfr_param{ctfmode} > 0) {
        $outpo2rmode = "ctf";
        if($target_refine){
          $target_refine_iter = 0;
        } 
    }

    # Run PO2R/PCTFR or PPFT to update particle origins and orientations
    if($auto_param{have_map} && $auto_param{generate_map} ne 'only') {
	--$auto_param{niter}; # Decrement number of iterations
	if($auto_param{mode} eq 'refine' && $outpo2rmode eq 'ctf') {
    	printf OUT "*** Updating defocus estimate\n\n";
    } elsif($auto_param{mode} eq 'refine' && $outpo2rmode eq 'mag') { 
    	printf OUT "*** Updating magnification\n\n";
    } else {
    	printf OUT "*** Updating origins and orientations\n\n";
    }

	# initialize common parameters
	if($auto_param{'adapt_angle'} eq 'yes') {
	  if($auto_param{mode} eq 'refine') {
	    $po2r_param{dangle} = 180./PI*$p3dr_param{res_min}/(2*2*$pixel_size*$auto_param{partrad});
	    $po2r_param{dangle} = $po2r_param{dangle} < 0.99*$prev_po2r_angle ? $po2r_param{dangle} : max(0.01,min(0.99*$prev_po2r_angle,$prev_po2r_angle-0.01));
	    $po2r_param{dangle} = sprintf("%.2f", $po2r_param{dangle});
	  } else {
	    $ppft_param{delta_theta} = 180./PI*$p3dr_param{res_min}/(2*$pixel_size*$auto_param{partrad});
	    $ppft_param{delta_theta} = ($ppft_param{delta_theta} < 0.99*$prev_ppft_angle ? $ppft_param{delta_theta} : 0.99*$prev_ppft_angle);
	    $ppft_param{delta_theta} = sprintf("%.2f", $ppft_param{delta_theta});
	    # ppft generates all the reprojections: it needs a limit
	    if($ppft_param{delta_theta} < 0.5) {
	      $ppft_param{delta_theta} = 0.5;
	    }
	  }	
	}
	
	my @pids;
	foreach my $subset (@{$ptlefiles{subsets}}) {
	  my $pid = fork();
	  die "fork failed: $!" unless defined $pid;
	  unless ($pid) {	  
	  printf OUT "Subset of particles: $subset\n";
	  my $suffix = $auto_param{"suffix_$subset"};
	  my $start_map = $auto_param{"start_map$suffix"};
	  my @subsetptlefiles = @{$ptlefiles{$subset}};

	  if($auto_param{mode} eq 'refine') {
	    my $template_map = $start_map;
	    my $background_map;
	    my $in_3d_name = $template_map;
	    unless(-r $in_3d_name) {
		  printf OUT "refine mode: unable to open $in_3d_name\n";
		  printf OUT "Exiting AUTO3DEM\n";
		  exit 1;
	    }
	    if($auto_param{flatten_map} || $target_refine_iter) {
	      if($target_refine_iter) {
		    $flatten_param{in_3d_name} = $po2r_param{"mask_map"};
		    $flatten_param{flag} = 1;
	      } else {
		    $flatten_param{in_3d_name} = $auto_param{"start_map"};
		    $flatten_param{flag} = 0;
		  }
		  unless(-r $flatten_param{in_3d_name}) {
		    printf OUT "refine mode: unable to open $flatten_param{in_3d_name}\n";
		    printf OUT "Exiting AUTO3DEM\n";
		    exit 1;
		  }
		  $flatten_param{out_file_name} = "flatten_map$suffix"; 

		  $flatten_param{mask_file_name} = "binary_mask$suffix"; 
		  $output_file = "flatten${suffix}_output_iter_${iter}";
		  system("flatten $flatten_param{in_3d_name} $flatten_param{mask_file_name} $flatten_param{flag} >&$output_file");
		  move $output_file, "$intfiles_dir/$output_file";
	      
		  my $falloff = $auto_param{flatten_falloff};
		  if($falloff < 0) {
		    $falloff = max(3,int($p3dr_param{res_min}/$pixel_size/2.0)+1);
		  }
		  $output_file = "maskmap${suffix}_output_iter_${iter}";
		  system("maskmap $flatten_param{mask_file_name} $template_map $flatten_param{out_file_name} $falloff >&$output_file");
		  move $output_file, "$intfiles_dir/$output_file";
		  
		  if($target_refine_iter) {
		    $output_file = "maskbkgmap${suffix}_output_iter_${iter}";
		    $background_map = "flatten_bkgmap$suffix"; 
		    system("maskmap $flatten_param{mask_file_name} $template_map $background_map $falloff 1 >&$output_file");
		    move $output_file, "$intfiles_dir/$output_file";
		  }
		  $in_3d_name = $flatten_param{out_file_name};
		  $template_map = $flatten_param{out_file_name};
		  unlink $flatten_param{mask_file_name};
	    }
	    if($auto_param{hollow_map} && !$target_refine_iter) {
		  $hollow_param{in_3d_name} = $template_map;
		  $hollow_param{out_file_name} = "hollow_map$suffix";
		  $in_3d_name = $hollow_param{out_file_name};
		  $input_file  = "hollow${suffix}_input_iter_${iter}";
		  $output_file = "hollow${suffix}_output_iter_${iter}";
		  make_program_input($input_file, 'pcut', \%hollow_param);
		  run_mpi_prog($$launch_cmd{$subset}, $pcut_param{bin}, $input_file,
			     $$pcut_cpu{$subset}, $output_file);
		  move $input_file,  "$intfiles_dir/$input_file";
		  move $output_file, "$intfiles_dir/$output_file";
	    }
	    # Run PCTFR to refine CTF parameters if the refine_ctf flag has been set.
	    if($auto_param{refine_ctf} && $pctfr_param{ctfmode} > 0) {
	      $input_file  = "pctfr${suffix}_input_iter_${iter}";
	      $output_file = "pctfr${suffix}_output_iter_${iter}";
	      $pctfr_param{in_3d_name}    = $in_3d_name;
	      make_program_input($input_file, 'pctfr', 
			       \%pctfr_param, \@subsetptlefiles);
	      run_mpi_prog($$launch_cmd{$subset}, $pctfr_param{bin}, $input_file,      
			 $$pctfr_cpu{$subset}, $output_file);
	      move $input_file,  "$intfiles_dir/$input_file";
	      move $output_file, "$intfiles_dir/$output_file";
	    } else {
	    # Run PO2R to refine orientation and origin parameters
	      $input_file  = "po2r${suffix}_input_iter_${iter}";
	      $output_file = "po2r${suffix}_output_iter_${iter}";
	      $po2r_param{in_3d_name} = $in_3d_name;
	      $po2r_param{bkg_3d_name} = $background_map;
	      $po2r_param{target_refine} = $target_refine_iter;

          make_program_input($input_file, 'po2r', 
			       \%po2r_param, \@subsetptlefiles);
	      run_mpi_prog($$launch_cmd{$subset}, $po2r_param{bin},  $input_file,       
			 $$po2r_cpu{$subset}, $output_file);
	      move $input_file,  "$intfiles_dir/$input_file";
	      move $output_file, "$intfiles_dir/$output_file";
	      if($target_refine_iter) {unlink $background_map};
	    }
	  } else {
	    $ppft_param{model_filename} = $start_map;
	    unless(-r $ppft_param{model_filename}) {
		printf OUT "PPFT unable to open $ppft_param{model_filename}\n";
		printf OUT "Exiting AUTO3DEM\n";
		exit 1;
	    }

	    if($auto_param{hollow_map}) {
		  $hollow_param{in_3d_name} = $ppft_param{model_filename};
		  $hollow_param{out_file_name} = "hollow_map$suffix";
		  $ppft_param{model_filename} = $hollow_param{out_file_name};
		  $input_file  = "hollow${suffix}_input_iter_${iter}";
		  $output_file = "hollow${suffix}_output_iter_${iter}";
		  make_program_input($input_file, 'pcut', \%hollow_param);
		  run_mpi_prog($$launch_cmd{$subset}, $pcut_param{bin}, $input_file,
			     $$pcut_cpu{$subset}, $output_file);
		  move $input_file,  "$intfiles_dir/$input_file";
		  move $output_file, "$intfiles_dir/$output_file";
	    }

	    $input_file  = "ppft_input_iter_${iter}$suffix";
	    $output_file = "ppft_output_iter_${iter}$suffix";
	    $ppft_param{pftrads_filename} = "ppft_iter_${iter}$suffix.rads";
	    $ppft_param{pftres1_filename} = "ppft_iter_${iter}$suffix.res1";
	    $ppft_param{pftres2_filename} = "ppft_iter_${iter}$suffix.res2";
	    if(-e $ppft_param{pftrads_filename}) {
		unlink $ppft_param{pftrads_filename};
	    }
	    if(-e $ppft_param{pftres1_filename}) {
		unlink $ppft_param{pftres1_filename};
	    }
	    if(-e $ppft_param{pftres2_filename}) {
		unlink $ppft_param{pftres2_filename};
	    }
	    make_program_input($input_file, 'ppft', 
			       \%ppft_param, \@subsetptlefiles);
	    run_mpi_prog($$launch_cmd{$subset}, $ppft_param{bin},  $input_file,       
			 $$ppft_cpu{$subset}, $output_file);

	    move $input_file,  "$intfiles_dir/$input_file";
	    move $output_file, "$intfiles_dir/$output_file";
	    move ($ppft_param{pftrads_filename}, 
		  "$intfiles_dir/$ppft_param{pftrads_filename}");
	    move ($ppft_param{pftres1_filename}, 
		  "$intfiles_dir/$ppft_param{pftres1_filename}");
	    move ($ppft_param{pftres2_filename}, 
		  "$intfiles_dir/$ppft_param{pftres2_filename}");
	  }
	  exit;
	  }
	  push @pids, $pid;
	  if($evod_parallel == 0) { 
	    waitpid($pid, 0);
	    my $pid_status = $?;
	    if ($pid_status) {exit 1};
	  }
	} # end loop over subsets
	
	if($evod_parallel == 1) {
	  for my $pid (@pids) {
	    waitpid($pid, 0);
	    my $pid_status = $?;
	    if ($pid_status) {exit 1};
	  }
	}

    # reset/update parameters after search
    if($auto_param{mode} eq 'refine') {
      if($auto_param{refine_ctf} && $pctfr_param{ctfmode} > 0) {
        # reset: only one iteration of ctf refinement done
        $auto_param{refine_ctf} = 0;
      } else {
        if($auto_param{'adapt_angle'} eq 'yes') {
          $prev_po2r_angle = $po2r_param{dangle};
        }
      }
    } else {
      if($auto_param{'adapt_angle'} eq 'yes') {
        $prev_ppft_angle = $ppft_param{delta_theta};
      }
    }
     
	# Increment the particle origin and orientation files
	incr_subsetptlefiles(\%ptlefiles);

	# Write out restart file
	my $restart_file = "$auto_param{outfile}_restart_${iter}a";
	$restart_file = "${startdir}/${restart_file}";
	$auto_param{have_map} = 0; # Upon restart, won't have map
	write_params($restart_file, \%auto_param, \%po2r_param,
		     \%p3dr_param, \%pcut_param, \%psf_param,
		     \%ppft_param, \%pctfr_param, \%ptlefiles);
    }

    my $res_lothresh_best = 99.99;
    my $res_hithresh_best = 99.99;
    my $res_cutoff_best = 99.99;
    my $outnselect = 0;

    unless($auto_param{generate_map} eq 'no') {
        print OUT "*** Evaluating selection criteria\n\n";
    
        my $reached_hithresh_best = 0;
        my $reached_lothresh_best = 0;
        my $reached_cutoff_best = 0;
        my $selection_best = 0;
        my @spatial_freq;
        my @fsc;

        # Make a copy of P3DR hash for creating even and odd maps and set
        # set the zero_fill to be one, and extend the resolution range.
        my %p3dr_param_oddeven = %p3dr_param;
        $p3dr_param_oddeven{zero_fill} = 1.0;
        $p3dr_param_oddeven{res_max} = ($p3dr_param_oddeven{res_max} < 3.0*$pixel_size) ? $p3dr_param_oddeven{res_max} : 3.0*$pixel_size;
        $p3dr_param_oddeven{res_max} = sprintf("%.2f", $p3dr_param_oddeven{res_max});
        $p3dr_param_oddeven{res_min} = $p3dr_param_oddeven{res_max};
        $psf_param{res_max} = $p3dr_param_oddeven{res_max};
        $psf_param{res_step} = int(2*$auto_param{partrad}*$pixel_size*(1./$psf_param{res_max}-1./$psf_param{res_min}));
        $psf_param{res_step} = ($psf_param{res_step} < 10) ? 10: $psf_param{res_step};
    
        my $gold_standard = $auto_param{gold_standard}; 

        my @selptlefiles;
        my $oddeven_flag;
        my $out_even_map;
        my $out_odd_map;

        # LOOP OVER SELECTION CRITERIA

        # If auto_param{estimate_res} is false, then NO resolution
        # estimation will be performed and selection criterion 0 will be
        # used to construct the map.
        # In the gold standard procedure, still the two half maps need
        # to be calculated

        my @jlist;
        if($auto_param{estimate_res}) {
	    my $nso = $auto_param{nselect_offset};
	    @jlist = (-$nso .. $nso);
        } elsif($gold_standard==1) {
	    $selection_best = 0;
	    @jlist = (0);
        } else {
	    $selection_best = 0;
	    @jlist = ();
        }

        foreach my $j (@jlist) {
	
	    my $suffix = "${iter}${map_suffix}_${j}";
	    $suffix =~ s/-/n/;
	
	    # Generate set parameters for reconstruction
	    if($gold_standard==1) {
	      @selptlefiles = @{$ptlefiles{even}};
	      $oddeven_flag = 0;
	      $out_even_map = $map_even_iter;
	    }
	    else {
	      @selptlefiles = @{$ptlefiles{all}};
	      $oddeven_flag = 2;
	      $out_even_map = "map_even";
	    }	

	    # Generate the filtered particle files
	    my ($nselect, $select_ref, $reject_ref) = 
	        gen_filtered_ptles(\@selptlefiles, \%auto_param, $j);

	    $pideven = fork();
        die("Cannot start even recostruction process!\n") unless defined $pideven;
	    unless ($pideven) {
	    # Run P3DR to generate even map
	    $input_file  = "p3dr_even_input_iter_$suffix";
	    $output_file = "p3dr_even_output_iter_$suffix";
	    $p3dr_param_oddeven{oddeven} = $oddeven_flag;
	    $p3dr_param_oddeven{out_file_name} = 'map_even';
	    make_program_input($input_file, 'p3dr', \%p3dr_param_oddeven, 
			   $select_ref);
	    run_mpi_prog($$launch_cmd{p3dreven}, $p3dr_param_oddeven{bin}, $input_file, 
		     $$p3dr_cpu{even}, $output_file);
	    move $input_file,  "$intfiles_dir/$input_file";
	    move $output_file, "$intfiles_dir/$output_file";
	
	    if($gold_standard==1) {
	      # Move _sel and _rej even files into $intfiles_dir directory
	      foreach my $selectfile (@{$select_ref}) {
	        move $selectfile, "$intfiles_dir/$selectfile";
	      }
	      foreach my $rejectfile (@{$reject_ref}) {
	        move $rejectfile, "$intfiles_dir/$rejectfile";
	      }
        }
	    exit;
	    }
	    if($evod_parallel == 0) { 
	      waitpid($pideven, 0);
	      my $pid_status = $?;
	      if ($pid_status) {exit 1};
	    }

	    if($gold_standard==1) {
	      @selptlefiles = @{$ptlefiles{odd}};
	      ($nselect, $select_ref, $reject_ref) = 
	        gen_filtered_ptles(\@selptlefiles, \%auto_param, $j);
	      $oddeven_flag = 0;
	      $out_odd_map = $map_odd_iter;
	    }
	    else {
	      @selptlefiles = @{$ptlefiles{all}};
	      $oddeven_flag = 1;
	      $out_odd_map = "map_odd";
	    }	

	    $pidodd = fork();
        die("Cannot start odd recostruction process!\n") unless defined $pidodd;
	    unless ($pidodd) {
	    # Run P3DR to generate odd map
	    $input_file  = "p3dr_odd_input_iter_$suffix";
	    $output_file = "p3dr_odd_output_iter_$suffix";
	    $p3dr_param_oddeven{oddeven} = $oddeven_flag;
	    $p3dr_param_oddeven{out_file_name}  = 'map_odd';
	    make_program_input($input_file, 'p3dr', \%p3dr_param_oddeven, 
			   $select_ref);
	    run_mpi_prog($$launch_cmd{p3drodd}, $p3dr_param_oddeven{bin}, $input_file,  
		     $$p3dr_cpu{odd}, $output_file);
	    move $input_file,  "$intfiles_dir/$input_file";
	    move $output_file, "$intfiles_dir/$output_file";
	
	    # Move _sel and _rej files into $intfiles_dir directory
	    foreach my $selectfile (@{$select_ref}) {
	        move $selectfile, "$intfiles_dir/$selectfile";
	    }
	    foreach my $rejectfile (@{$reject_ref}) {
	        move $rejectfile, "$intfiles_dir/$rejectfile";
	    }
	    exit;
	    }
	    if($evod_parallel == 1) { 
	      waitpid($pideven, 0);
	      my $pid_status = $?;
	      if ($pid_status) {exit 1};
	    }
	    waitpid($pidodd, 0);
	    my $pid_status = $?;
	    if ($pid_status) {exit 1};

	    if($auto_param{estimate_res}) {
	    $pideven = fork();
        die("fork failed: $!\n") unless defined $pideven;
	    unless ($pideven) {
	    # Run PCUT to cut the even map
	    $input_file  = "pcut_even_input_iter_$suffix";
	    $output_file = "pcut_even_output_iter_$suffix";
	    $pcut_param{in_3d_name}    = 'map_even';
	    $pcut_param{out_file_name} = "map_even_cut";
	    make_program_input($input_file, 'pcut', \%pcut_param);
	    run_mpi_prog($$launch_cmd{even}, $pcut_param{bin}, $input_file, 
		     $$pcut_cpu{even}, $output_file);
	    move $input_file,  "$intfiles_dir/$input_file";
	    move $output_file, "$intfiles_dir/$output_file";
	    exit;
	    }
	    if($evod_parallel == 0) { 
	      waitpid($pideven, 0);
	    }
	
	    $pidodd = fork();
        die("fork failed: $!\n") unless defined $pidodd;
	    unless ($pidodd) {
	    # Run PCUT to cut the odd map
	    $input_file  = "pcut_odd_input_iter_$suffix";
	    $output_file = "pcut_odd_output_iter_$suffix";
	    $pcut_param{in_3d_name}     = 'map_odd';
	    $pcut_param{out_file_name}  = "map_odd_cut";
	    make_program_input($input_file, 'pcut', \%pcut_param);
	    run_mpi_prog($$launch_cmd{odd}, $pcut_param{bin}, $input_file,  
		     $$pcut_cpu{odd}, $output_file);
	    move $input_file,  "$intfiles_dir/$input_file";
	    move $output_file, "$intfiles_dir/$output_file";
	    exit;
	    }
	    if($evod_parallel == 1) { 
	      waitpid($pideven, 0);
	    }
	    waitpid($pidodd, 0);
	
	    # Run PSF to generate FSC curve
	    $input_file  = "psf_input_iter_$suffix";
	    $output_file = "psf_output_iter_$suffix";
	    $psf_param{fileeven} = "map_even_cut";
	    $psf_param{fileodd}  = "map_odd_cut";
	    make_program_input($input_file, 'psf', \%psf_param);
	    run_mpi_prog($$launch_cmd{all}, $psf_param{bin},  $input_file,       
		     $$psf_cpu{all}, $output_file);

	    # Parse the PSF output file to determine current resolution and
	    # high resolution to be used in next iteration

	    my ($res_hithresh, $res_lothresh, $res_cutoff, $reached_hithresh,
	        $reached_lothresh, $reached_cutoff, $spatial_freq_ref, $fsc_ref) =
	        psf_parse($auto_param{fsc_hithresh}, $auto_param{fsc_lothresh},
		      $auto_param{fsc_cutoff}, $output_file);
	
	    # Determine whether this selection criterion results in
	    # the best resolution
	    if($res_hithresh < $res_hithresh_best) {
	        $res_hithresh_best = $res_hithresh;
	        $res_lothresh_best = $res_lothresh;
	        $res_cutoff_best = $res_cutoff;
	        $selection_best = $j;
	        $reached_hithresh_best = $reached_hithresh;
	        $reached_lothresh_best = $reached_lothresh;
	        $reached_cutoff_best = $reached_cutoff;
	        @spatial_freq = @{$spatial_freq_ref};
	        @fsc = @{$fsc_ref};
	        copy($output_file, "$auto_param{outfile}_FSC_iter_${iter}${map_suffix}");
	        if($gold_standard and $auto_param{new_ptles}==0) {
	          move "map_even", "$out_even_map";
	          move "map_odd", "$out_odd_map";
	        }
	    }

	    move $input_file,  "$intfiles_dir/$input_file";
	    move $output_file, "$intfiles_dir/$output_file";
	
	    # Cleanup temporary map files
	    unlink 'map_even', 'map_odd', 'map_even_cut', 'map_odd_cut';
        } else {
          # rename maps from single selection loop to be references for next iter
          unless ($auto_param{new_ptles}) {
	        move "map_even", "$out_even_map";
	        move "map_odd", "$out_odd_map";
	      }
        }
        } # end loop over selection criteria in jlist

        if($auto_param{estimate_res}) {
	      # Construct xmgrace input file and PNG (if possible)
	      my $fsctitle = "$auto_param{outfile} iter $iter${map_suffix}";
	      my $fsctxt = "$auto_param{outfile}_FSC_iter_${iter}${map_suffix}";
	      my $fscpng = "${fsctxt}.png";
	      my $fscagr = "${fsctxt}.agr";
	      make_grace_fscfile(\@spatial_freq, \@fsc, $fsctitle, $iter, $fscagr);
	      if(defined $gracebat) {
	        system("$gracebat -printfile $fscpng -hdevice PNG $fscagr");
	      }
	
	      # Send notification that iteration has completed and attach FSC curve
	      if(defined $auto_param{recipient}) {
	        my $subj = "FSC curve - iteration $iter ${map_suffix}";
	        my @attach = ("$fsctxt");
	        if(defined $gracebat) {
		      push(@attach, "$fscpng");
	        }
	        my $msgfile = "message.txt";
	        my $currentdir = getcwd();
	        open(MSG, ">$msgfile");
	        print MSG "Completed $auto_param{outfile} iteration $iter\n";
	        print MSG "FSC data attached\n";
	        close(MSG);
	        $msgfile = "$currentdir/$msgfile";
	        sendmail($msgfile, $subj, $auto_param{recipient}, @attach);
	        unlink $msgfile;
	      }

	      print OUT "*** Construct map using selection criterion ";
	      print OUT "$selection_best\n";

	      # Adjust resolutions based on the best resolution estimate
          update_res(\%po2r_param, \%p3dr_param, \%psf_param,
	       \%ppft_param, \%pctfr_param, \%auto_param,
	       $pixel_size, $res_lothresh_best,
	       $reached_lothresh_best, $res_hithresh_best,
	       $reached_hithresh_best, $res_cutoff_best,
	       $reached_cutoff_best, $auto_param{freeze_res}); 
        }

        # Generate the filtered particle files using best selection criterion
        # Warning: in gold standard procedure this selection
        #          will be slightly different than the one used for estimating
        #          the resolution
        my ($nselect, $select_ref, $reject_ref) = 
	      gen_filtered_ptles(\@{$ptlefiles{all}}, \%auto_param, $selection_best);
        $outnselect = $nselect;
        
        # If we have new particles that we are refining relative to an
        # existing higher resolution map, we do not need to construct a
        # new map.
        unless($auto_param{new_ptles}) {

	      # Write out FSC data file to be read by P3DR
	      my $fsc_file_name;
	      if($auto_param{noise_suppression}) {
	        my $fsc_count = @fsc;
	        $fsc_file_name = "fscfile_${iter}${map_suffix}";
	        open(FSC, ">$fsc_file_name");
	        print FSC "$fsc_count\n";
	        for(my $k=0; $k<$fsc_count; $k++) {
		      if($fsc[$k] < 0) {
		        printf FSC "%7.4f\t%7.4f\n", $spatial_freq[$k], 0.0;
		      } else {
		        printf FSC "%7.4f\t%7.4f\n", $spatial_freq[$k], $fsc[$k];
		      }
	        }
	        close(FSC);
	        $p3dr_param{fsc_file_name} = $fsc_file_name;
	      }
    
	      # Run P3DR to generate full map using filtered particle files
	      $p3dr_param{out_file_name} = "$map_iter";
	      $input_file  = "p3dr_input_iter_${iter}${map_suffix}";
	      $output_file = "p3dr_output_iter_${iter}${map_suffix}";
	      $p3dr_param{oddeven} = 0;
	      make_program_input($input_file, 'p3dr', \%p3dr_param, $select_ref);
	      run_mpi_prog($$launch_cmd{p3drall}, $p3dr_param{bin}, $input_file,      
		     $$p3dr_cpu{all}, $output_file);
	      delete $p3dr_param{fsc_file_name};

	      move $input_file,     "$intfiles_dir/$input_file";
	      move $output_file,    "$intfiles_dir/$output_file";
	      if($auto_param{noise_suppression}) {
	        move $fsc_file_name,  "$intfiles_dir/$fsc_file_name";
	      }

	    }
	    elsif($gold_standard==1 and $iter!=$last_iter) {
          # in gold standard procedure calculate average between two maps
          # to determine radial constraints
          $input_file = "avg_input";
          $output_file = "avg_output";
          open(INP, ">$input_file") || die "Unable to open file $input_file\n";
          print INP "2\n";
          print INP "$map_even_iter\n";
          print INP "$map_odd_iter\n";
          print INP "$map_iter\n";
          close(INP);
          system("avemaps <$input_file > $output_file");
          unlink $input_file;
          unlink $output_file;
	    }
	    
	    # if new particles added, at each iteration a map is not generated,
	    # therefore its radial average is not calculated, and new radial
	    # parameters are not estimated
	    unless ($auto_param{new_ptles}) {
	      # Set start_map to be used for next iteration or run
	      $auto_param{start_map} = "$map_iter";
	      $auto_param{start_map_even} = "$map_even_iter";
	      $auto_param{start_map_odd} = "$map_odd_iter";
        
          # Generate radial density profile
          my $rad_profile = "$auto_param{outfile}_radavg_${iter}${map_suffix}";
          unlink $rad_profile;
          system("imgstats $map_iter > $rad_profile");
    
          # Parse radial density profile to update inner and outer radii of
          # capsid. Use these to set (annulus_low, annulus high) for next
          # run of PPFT and (in_rad, out_rad) for PCUT

          my ($cut_lo, $cut_hi, $ann_lo, $ann_hi, $ann_peak, $capsid_score) = rad_profile_parse($rad_profile);
          print OUT "   capsid score  = $capsid_score\n";
        
          if ($auto_param{adapt_angle} eq 'yes' && $ann_peak > $auto_param{boxrad}/3.0) {
              $auto_param{partrad} = $ann_peak;
          }

          if($auto_param{hollow_map} && $auto_param{hollow_auto}) {
            $auto_param{hollow_in_rad}  = $ann_lo;
            $auto_param{hollow_out_rad} = $ann_hi;
            $hollow_param{in_rad}       = $auto_param{hollow_in_rad};
            $hollow_param{out_rad}      = $auto_param{hollow_out_rad};
          }
        
          unless($auto_param{freeze_annulus}) {
	
	        # Make sure that ann_lo >= boxrad/3
	        if($ann_lo < int($auto_param{boxrad}/3.0)) {
	          $ann_lo = int($auto_param{boxrad}/3.0);
	        }
	        if($cut_lo < int($auto_param{boxrad}/3.0)) {
	          $cut_lo = int($auto_param{boxrad}/3.0);
	        }
	
	        $ppft_param{annulus_low} = $ann_lo;
	        $ppft_param{annulus_high} = $ann_hi;
	        $pcut_param{in_rad} = $cut_lo;
	        $pcut_param{out_rad} = $cut_hi;
	        print OUT "   PPFT annulus_low changed to  $ann_lo\n";
	        print OUT "   PPFT annulus_high changed to $ann_hi\n";
	        print OUT "   PCUT in_rad changed to       $cut_lo\n";
	        print OUT "   PCUT out_rad changed to      $cut_hi\n";
	
	        if($auto_param{rmc} == 2) { # New RMC calculations
	          $auto_param{hollow_out_rad} = $ann_hi;
	          $hollow_param{out_rad} = $ann_hi;
	          print OUT "   hollow_out_rad changed to    $ann_hi\n";
	        }
	      }  
        }
    
        # Move _sel files into $intfiles_dir directory
        foreach my $selectfile (@{$select_ref}) {
	    move $selectfile, "$intfiles_dir/$selectfile";
        }
        foreach my $rejectfile (@{$reject_ref}) {
	    move $rejectfile, "$intfiles_dir/$rejectfile";
        }

        # Gather data for the summary information
        unless($reached_cutoff_best) {
	    $fscthresh_flag = "*";
        }

        # Add map file to list
        unless ($auto_param{new_ptles}) {
          push(@mapfiles, $map_iter);
          if($gold_standard) {
            push(@eomapfiles, $map_even_iter);
            push(@eomapfiles, $map_odd_iter);
          }
        }

        # Add estimated resolution to list
        unshift(@estres, $res_cutoff_best);

        # Test progress and switch from search to refine mode if
        #   (1) Current mode is search
        #   (2) Resolution improved by less than 0.5A
        #   (3) switch_mode flag is set
        #   (4) PPFT bin_factor == 1 or bin_reduce not set

        if(defined $estres[0] && defined $estres[1] &&
          $auto_param{mode} eq 'search' && 
          ($estres[1]-$estres[0]) < 0.5 &&
          $auto_param{switch_mode} &&
          ($ppft_param{bin_factor} == 1 || !$auto_param{bin_reduce})) {
	    $auto_param{mode} = 'refine';
        }
    
        # Test progress and reduce bin factor if
        #   (1) Current mode is search
        #   (2) Resolution improved by less than 0.25A
        #   (3) PPFT bin_factor > 1
        #   (4) bin_reduce flag set

        if(defined $estres[0] && defined $estres[1] &&
          $auto_param{mode} eq 'search' && 
          ($estres[1]-$estres[0]) < 0.25 &&
          $ppft_param{bin_factor} > 1 && $auto_param{bin_reduce}) {
	    --$ppft_param{bin_factor};
        }

        # Write out restart file
        my $restart_file = "$auto_param{outfile}_restart_${iter}b";
        $restart_file = "${startdir}/${restart_file}";
        $auto_param{have_map} = 1; # Upon restart, will have map
        write_params($restart_file, \%auto_param, \%po2r_param,
		 \%p3dr_param, \%pcut_param, \%psf_param,
		 \%ppft_param, \%pctfr_param, \%ptlefiles);


        # Terminate auto3dem run only if capsid score fails to improve by
        # at least 2%. Note that this affect only this calculation and has
        # no effect on any other RMC calculations that may be running.

        if($auto_param{term_search} && $capsid_score < 1.02*$prev_capsid_score) {
	    print OUT "\n--- Terminating auto3dem early ---\n";
	    print OUT "auto term_search flag set and capsid score\n";
	    print OUT "failed to improve by at least 2%\n\n";
	    last;
        }

        # Save capsid score for testing in next iteration
        $prev_capsid_score = $capsid_score;
    

        # Quit early if flag set and resolution is very 'high'
        # Also create RMC_DONE so no other RMC calculations are launched

        if($res_hithresh_best <= $psf_param{res_max} &&
          $auto_param{quit_early}) {
	    print "\n--- Terminating auto3dem early ---\n";
	    print "Starting model successfully calculated\n\n"; 
	    print "Note that this message should only be encountered when\n";
	    print "using the random model method. If you see this message\n";
	    print "during an ordinary auto3dem run, remove the line\n";
	    print "auto \"quit_early 1\" from your auto3dem master file\n\n";
	    print OUT "\n--- Terminating auto3dem early ---\n";
	    print OUT "Starting model successfully calculated\n\n"; 
	    print OUT "Note that this message should only be encountered when\n";
	    print OUT "using the random model method. If you see this message\n";
	    print OUT "during an ordinary auto3dem run, remove the line\n";
	    print OUT "auto \"quit_early 1\" from your auto3dem master file\n\n";
	    chdir($startdir); # Make sure file written in correct place
	    open(DONE, ">RMC_DONE");
	    print DONE "DONE - starting model successfully calculated\n";
	    close(DONE);
	    chdir($auto_param{rundir});
	    last;
        }
    }  # end unless(generate_map==no)

    # Write out summary information

    # Get count of total number of ptles and micrographs, plus range
    # of defocus levels in parameter files
    my ($ntotal, $nmicro, $defmin, $defmax) = (0, 0, 99.99, -99.99);
    my %micro_read;
    foreach my $file (@{$ptlefiles{all}}) {
	my ($np, $nm, $dm, $dx, @stacknames) = get_paramfile_info($file);
	$ntotal += $np;
	$nmicro += $nm;
	if($dm < $defmin) {
	    $defmin = $dm;
	}
	if($dx > $defmax) {
	    $defmax = $dx;
	}
	# check actual number of micrographs
	if($auto_param{gold_standard}) {
	   foreach my $st (@stacknames) {
	      unless ($micro_read{$st}) {
	         $micro_read{$st}=1;
	      }
	   }
	}
    }
	if($auto_param{gold_standard}) {
	   $nmicro = scalar keys %micro_read;
	}   
    
    my $delta_angle;
    if($auto_param{generate_map} eq 'only') {
        $delta_angle = "0.00"; 
    } elsif($outmode eq "refine") {
        if($outpo2rmode eq 'local') {
            $delta_angle = $po2r_param{dangle};
        } elsif ($outpo2rmode eq 'global') {
	        $delta_angle = $po2r_param{gangle};
	    } elsif ($outpo2rmode eq 'mag') {
	        $delta_angle = "0.00";
	    } elsif ($outpo2rmode eq 'ticos_equiv') {
	        $delta_angle = "0.00";
	    } elsif ($outpo2rmode eq 'ctf') {
	        $delta_angle = $pctfr_param{dangle};
	    }
    } else {
	    $delta_angle = $ppft_param{delta_theta};
    }

    my $time_iter_end= time();
    my $time_iter = $time_iter_end - $time_iter_start;

    if($auto_param{generate_map} eq 'only') {
        $outmode = 'map'; 
    } elsif($outmode eq 'search') {
	    $outmode = "s($ppft_param{bin_factor})";
    } elsif ($outmode eq 'refine') {
	    if($outpo2rmode eq 'local') {
	      $outmode = 'r(l)';
	    } elsif ($outpo2rmode eq 'mag') {
	      $outmode = 'r(m)';
        } elsif ($outpo2rmode eq 'global') {
	      $outmode = 'r(g)';
	    } elsif ($outpo2rmode eq 'ticos_equiv') {
	      $outmode = 'r(i)';
	    } elsif ($outpo2rmode eq 'ctf') {
	      $outmode = 'r(c)';
	    } else {
	      $outmode = "----";
	    }
    } else {
	    $outmode = "----";
    }

    printf SUM ("%3d %4s ",$iter, $outmode);
    if($auto_param{generate_map} eq 'no') {
        printf SUM ("  ---  %6.2f    ---    --- %5d %3d    --- ",
		$delta_angle, $time_iter, $ncpu);
    } else {
        printf SUM ("%5.2f%1s %6.2f %6.2f %6.2f %5d %3d %6d ",
		$res_cutoff_best, $fscthresh_flag, $delta_angle,
		$p3dr_param{res_min}, $p3dr_param{res_max},
		$time_iter, $ncpu, $outnselect);
    }
    printf SUM ("%6d %3d %4.2f-%4.2f\n", $ntotal, $nmicro, $defmin, $defmax);



}

# Delete intermediate map files if flag set
if($auto_param{delete_maps}) {
  pop @mapfiles; # Pop the last file added off the list
  foreach my $map (@mapfiles) {
	unlink $map;
  }
  if($auto_param{gold_standard}) {
    pop @eomapfiles; # Pop the last two files added off the list
    pop @eomapfiles; 
    foreach my $map (@eomapfiles) {
	  unlink $map;
    }
  }
}

# Get rid of any hollow_map files
unlink "flatten_map";
unlink "flatten_map$auto_param{suffix_even}";
unlink "flatten_map$auto_param{suffix_odd}";
unlink "hollow_map";
unlink "hollow_map$auto_param{suffix_even}";
unlink "hollow_map$auto_param{suffix_odd}";

# Get rid of stdin_temp file (generated by PPFT)
unlink "stdin_temp";

if (@hostfiles) {
  #delete all the files listed in the array
  foreach (@hostfiles) {
    unlink $_;
  }
}

# Return to starting directory
chdir($startdir);

# Move restart files into their own directory
my $restart_dir = "$auto_param{outfile}_RESTARTS";
mkdir $restart_dir;
foreach my $file (<$auto_param{outfile}_restart*>) {
    move $file, $restart_dir;
}

# Restore number of iterations to original value
$auto_param{niter} = $niter_original;

# the name of the continue file indicates the next iteration to run
my $continue_file = "$auto_param{outfile}_continue";

# adjust parameters
if($auto_param{generate_map} eq 'only') {
   $auto_param{generate_map} = 'yes';
   $auto_param{map_suffix} = 'none';
   $continue_file = "${continue_file}_$auto_param{iter_start}${map_suffix}";
}
elsif($auto_param{generate_map} eq 'no') {
   $auto_param{generate_map} = 'only';
   $continue_file = "${continue_file}_rec${last_iter}";
}
elsif($auto_param{generate_map} eq 'yes') {
   $continue_file = "${continue_file}_$auto_param{iter_start}";
}

write_params($continue_file, \%auto_param, \%po2r_param,
   \%p3dr_param, \%pcut_param, \%psf_param,
   \%ppft_param, \%pctfr_param, \%ptlefiles);

my $time_stop = time();
my $time_elapsed = $time_stop - $time_start;
print OUT "Auto3dem finished\n";
print OUT "   Total run time = $time_elapsed seconds\n";

close(OUT);
close(SUM);
