use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

sub init_cpus {

    # Purpose: set the number of cpus assigned to each program.
    # By default they are read from the input parameter file, but they are
    # modified if even and odd processing are set to run concurrently.
    #
    # $PROG_param{max_cpu} is the upper limit
    #
    #
    # Input arguments:
    #   $ncpu            - max number of cpus allowed
    #   $evod_parallel - flag to run even and odd processes concurrently
    #   $po2r_param      - reference to hash of PO2R input parameters
    #   $p3dr_param      - reference to hash of P3DR input parameters
    #   $pcut_param      - reference to hash of PCUT input parameters
    #   $psf_param       - reference to hash of PSF input parameters
    #   $ppft_param      - reference to hash of PSF input parameters
    #   $pctfr_param     - reference to hash of PCTFR input parameters
    #
    # Return values:
    #   $po2r_cpu   - reference to hash of cpus for PO2R
    #   $p3dr_cpu   - reference to hash of cpus for P3DR
    #   $pcut_cpu   - reference to hash of cpus for PCUT
    #   $psf_cpu    - reference to hash of cpus for PSF
    #   $ppft_cpu   - reference to hash of cpus for PSF
    #   $pctfr_cpu  - reference to hash of cpus for PCTFR

    my ($ncpu, $evod_parallel, $po2r_param, $p3dr_param, $pcut_param, $psf_param,
    $ppft_param, $pctfr_param) = @_;

    my (%po2r_cpu, %p3dr_cpu, %pcut_cpu, %psf_cpu, %ppft_cpu, %pctfr_cpu);

    if ($ncpu < 2 && $evod_parallel == 1) {
        print "[init_cpus.pm] Input parameters ncpu ($ncpu) and concurrence requested ($evod_parallel) incompatible!\n";
	    exit;
	}

    # Set the number of CPUs to be used by each program
    # ($PROG_param{max_cpu} is the upper limit)
    $po2r_cpu{all}  = ($$po2r_param{max_cpu}  < $ncpu) ? $$po2r_param{max_cpu}  : $ncpu;
    $p3dr_cpu{all}  = ($$p3dr_param{max_cpu}  < $ncpu) ? $$p3dr_param{max_cpu}  : $ncpu;
    $pcut_cpu{all}  = ($$pcut_param{max_cpu}  < $ncpu) ? $$pcut_param{max_cpu}  : $ncpu;
    $psf_cpu{all}   = ($$psf_param{max_cpu}   < $ncpu) ? $$psf_param{max_cpu}   : $ncpu;
    $ppft_cpu{all}  = ($$ppft_param{max_cpu}  < $ncpu) ? $$ppft_param{max_cpu}  : $ncpu;
    $pctfr_cpu{all} = ($$pctfr_param{max_cpu} < $ncpu) ? $$pctfr_param{max_cpu} : $ncpu;

    $po2r_cpu{even}  = $po2r_cpu{odd}  = $po2r_cpu{all};
    $p3dr_cpu{even}  = $p3dr_cpu{odd}  = $p3dr_cpu{all};
    $pcut_cpu{even}  = $pcut_cpu{odd}  = $pcut_cpu{all};
    $psf_cpu{even}   = $psf_cpu{odd}   = $psf_cpu{all};
    $ppft_cpu{even}  = $ppft_cpu{odd}  = $ppft_cpu{all};
    $pctfr_cpu{even} = $pctfr_cpu{odd} = $pctfr_cpu{all};

    if ($evod_parallel == 1 )  {
        if (2*$po2r_cpu{all} > $ncpu) {
            $po2r_cpu{even} = int($ncpu/2);
            $po2r_cpu{odd} = $ncpu - $po2r_cpu{even};
        }
        if (2*$p3dr_cpu{all} > $ncpu) {
            $p3dr_cpu{even} = int($ncpu/2);
            $p3dr_cpu{odd} = $ncpu - $p3dr_cpu{even};
        }
        if (2*$pcut_cpu{all} > $ncpu) {
            $pcut_cpu{even} = int($ncpu/2);
            $pcut_cpu{odd} = $ncpu - $pcut_cpu{even};
        }
        if (2*$psf_cpu{all} > $ncpu) {
            $psf_cpu{even} = int($ncpu/2);
            $psf_cpu{odd} = $ncpu - $psf_cpu{even};
        }
        if (2*$ppft_cpu{all} > $ncpu) {
            $ppft_cpu{even} = int($ncpu/2);
            $ppft_cpu{odd} = $ncpu - $ppft_cpu{even};
        }
        if (2*$pctfr_cpu{all} > $ncpu) {
            $pctfr_cpu{even} = int($ncpu/2);
            $pctfr_cpu{odd} = $ncpu - $pctfr_cpu{even};
        }
    }

    return (\%po2r_cpu, \%p3dr_cpu, \%pcut_cpu, \%psf_cpu, \%ppft_cpu, \%pctfr_cpu);
}
1;

