use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

sub init_launch_cmd {

    # Set the command used to launch serial or parallel jobs.
    # If parallel jobs, the variable "NCPU" will 
    # be replaced with the actual number of CPUs when job is launched.
    # If even and odd processing are required to run concurrently,  
    # generate two hostfiles starting from the input hostfile,
    # which is split alternating the fields
    # (i.e. odd lines to hostfile for odd processing,
    #       even lines to hostfile for even processing)
    #
    # NOTE: the way the nodefile is split into even and odd hostlist is
    #       consistent with the number of cpus assigned in init_cpus
    #
    #
    # Input arguments:
    #   $mode           - parallel/serial mode flag
    #   $nodedir        - directory where to write the host files
    #   $nodefile       - file containing list of nodes
    #   $evod_parallel  - flag to indicate if even/odd processing done in parallel
    #   $p3drcpu_sparse - flag to indicate if cpus for p3dr to be spread among different nodes
    #
    # Return values:
    #   $hostfiles2del - array of host files to DELETE
    #   $launch_cmd    - reference to hash of command lines for each different task

    my ($mode, $p3dr_cpu, $nodedir, $nodefile, $evod_parallel, $p3drcpu_sparse) = @_;

    my $has_nodefile = 0;    # nodefile was either retrieved or given by user
    my %launch_cmd;
    my %hostfiles;
    my @hostfiles2del;

    # Default serial behavior
    $launch_cmd{all} = $launch_cmd{even} = $launch_cmd{odd} = "";
    $launch_cmd{p3drall} = $launch_cmd{p3dreven} = $launch_cmd{p3drodd} = "";
    $hostfiles{all} = $hostfiles{even} = $hostfiles{odd} = $nodefile;
    $hostfiles{p3drall} = $hostfiles{p3dreven} = $hostfiles{p3drodd} = $nodefile;

    if($mode eq 'parallel') {
        my $basic_cmd;
        if (`which mpirun` =~ /\/mpirun/) {
            $basic_cmd = "mpirun -np NCPU";
            if(defined $nodefile && $nodefile !~ /^none$/i) {
                $has_nodefile = 1;
            }
        } elsif (`which aprun` =~ /\/aprun/) {
            $basic_cmd = "aprun -n NCPU";
        } else {
            print "\n\n----- Exiting AUTO3DEM -----\n\n";
            print "Cannot find mpirun or aprun commands\n";
            print "\n\n----- Exiting AUTO3DEM -----\n\n";
    
           	print OUT "\n\n----- Exiting AUTO3DEM -----\n\n";
            print OUT "Cannot find mpirun or aprun commands\n";
            print OUT "\n\n----- Exiting AUTO3DEM -----\n\n";
            exit(0);
        }
    
        if ($has_nodefile == 1) {
            # read all nodes in nodefile
            open(NF, "< $nodefile") or die "Failed to open file: $!\n";
            chomp(my @hostlist = <NF>);
            close(NF);
            # reorder list in nodefile to iterate over nodes (roundrobin)
            my (@roundbordered_hostlist) = reorder_roundrobin(@hostlist);

            # reorder list in nodefile by processor affinity
            my @praffordered_hostlist = reorder_procaffinity(@hostlist);

            if ($p3drcpu_sparse == 1) {
                # create hostfiles optimized for p3dr
                $hostfiles{p3drall} = "${nodedir}/_a3dm_p3drall_hostlist";
                push(@hostfiles2del, $hostfiles{p3drall});

                open(OF, ">$hostfiles{p3drall}") || die "Unable to generate file $hostfiles{p3drall}\n";; 
       	        for my $i (0 .. ($$p3dr_cpu{all}-1)) {
                    print OF "$roundbordered_hostlist[$i]\n";
        	    }
      	        close(OF);
      	    }

            if ($evod_parallel == 0) {
                # same hostfile as for reconstructing from all particles
                $hostfiles{p3dreven} = $hostfiles{p3drall};
                $hostfiles{p3drodd} = $hostfiles{p3drall};
            } else {
                $hostfiles{even} = "${nodedir}/_a3dm_even_hostlist";
                $hostfiles{odd} = "${nodedir}/_a3dm_odd_hostlist";
                push(@hostfiles2del, $hostfiles{even});
                push(@hostfiles2del, $hostfiles{odd});

                my ($evenlist, $oddlist) = eosplit_hostlist(@praffordered_hostlist);
                my @evenlist = @$evenlist;
                my @oddlist = @$oddlist;
                
                open(OOUT, ">$hostfiles{odd}") || die "Unable to generate file $hostfiles{odd}\n";; 
                foreach (@oddlist) {
                    print OOUT "$_\n";
                }
                close(OOUT);
                open(EOUT, ">$hostfiles{even}") || die "Unable to generate file $hostfiles{even}\n";; 
                foreach (@evenlist) {
                    print EOUT "$_\n";
                }
                close(EOUT);

                # repeat procedure for the reconstruction-related hostfiles
                $hostfiles{p3dreven} = "${nodedir}/_a3dm_p3dreven_hostlist";
                $hostfiles{p3drodd} = "${nodedir}/_a3dm_p3drodd_hostlist";
                push(@hostfiles2del, $hostfiles{p3dreven});
                push(@hostfiles2del, $hostfiles{p3drodd});

                my @eohostlist;
                if ($p3drcpu_sparse == 1) {
                    @eohostlist = @roundbordered_hostlist;
                } else {
                    @eohostlist = @praffordered_hostlist;
                }

                open(OOUT, ">$hostfiles{p3drodd}") || die "Unable to generate file $hostfiles{odd}\n";; 
       	        for my $i (0 .. ($$p3dr_cpu{odd}-1)) {
                    print OOUT "$eohostlist[$i]\n";
        	    }
                close(OOUT);
                open(EOUT, ">$hostfiles{p3dreven}") || die "Unable to generate file $hostfiles{even}\n";; 
       	        for my $i ($$p3dr_cpu{odd} .. ($$p3dr_cpu{odd}+$$p3dr_cpu{even}-1)) {
                    print EOUT "$eohostlist[$i]\n";
        	    }
                close(EOUT);
                
            }
	        $launch_cmd{all} = "$basic_cmd -machinefile $hostfiles{all}";
	        $launch_cmd{even} = "$basic_cmd -machinefile $hostfiles{even}";
	        $launch_cmd{odd} = "$basic_cmd -machinefile $hostfiles{odd}";
	        $launch_cmd{p3drall} = "$basic_cmd -machinefile $hostfiles{p3drall}";
	        $launch_cmd{p3dreven} = "$basic_cmd -machinefile $hostfiles{p3dreven}";
	        $launch_cmd{p3drodd} = "$basic_cmd -machinefile $hostfiles{p3drodd}";
	    } else {
	        $launch_cmd{odd} = $launch_cmd{even} = $launch_cmd{all} = "$basic_cmd";
	        $launch_cmd{p3drodd} = $launch_cmd{p3dreven} = $launch_cmd{p3drall} = "$basic_cmd";
	    }
	}
    return (\@hostfiles2del,\%launch_cmd);
}

sub reorder_roundrobin{

    my (@inlist) = @_;
    
    my @outlist = ();
    my @remaind = @inlist;
    while (@remaind) {
        my ($uniqs,$remaind) = extract_uniq(@remaind);
        @remaind = @$remaind;
        push(@outlist, @$uniqs);
    }
    return (@outlist);
}

sub reorder_procaffinity{

    my (@inlist) = @_;
    
    my @outlist = ();
    my @remaind = @inlist;
    while (@remaind) {
        my ($same,$remaind) = extract_same(@remaind);
        @remaind = @$remaind;
        push(@outlist, @$same);
    }
    return (@outlist);
}

sub extract_uniq{

    my (@inputarr) = @_;
    my %seen;
    my @uniq;
    my @remaind;
    
    foreach my $item (@inputarr) {
        chomp($item);
        unless ($seen{$item}) {
            $seen{$item} = 1;
            push(@uniq, $item);
        } else {
            push(@remaind, $item);
        }
    }
    return (\@uniq,\@remaind)
}

sub extract_same{

    my (@inputarr) = @_;
    my %chosen;
    my @same;
    my @remaind;
    
    if (@inputarr) {
        $chosen{$inputarr[0]} = 1;
    }
    foreach my $item (@inputarr) {
        chomp($item);
        if ($chosen{$item}) {
            push(@same, $item);
        } else {
            push(@remaind, $item);
        }
    }
    return (\@same,\@remaind)
}

sub eosplit_hostlist{

    my (@inputlist) = @_;

    my $n = scalar @inputlist;
    if ($n < 2) {
        print "ERROR[init_launch_cmd:split_hostlist]: trying to split list with $n items!\n";
        exit;
    }
    my $half = $n - $n/2; #number of hosts assigned to odd process
    
    my @evenlist = ();
    my @oddlist = ();
    for my $i (0 .. ($half-1)) {
       push(@oddlist,$inputlist[$i]);
    }  
    for my $i ($half .. $#inputlist) {
       push(@evenlist,$inputlist[$i]);
    }  

    return (\@evenlist,\@oddlist);
}
1;

