#!/usr/bin/env python
"""
Module with parameter data structures
"""
from math import sin, cos, radians

class ParamFile:

    def __init__(self, param_file =''):
        self.paramFilename = param_file
        self.Stacks = []
        if param_file:
            self.ReadFile(param_file)
    
    def __str__(self):
        description = "\nParameter file: %s\n"%self.paramFilename
        description += "Number of stacks: %d\n"%self.NumberOfStacks()
        npart = 0
        description += "Total number of particles: %d\n"%self.NumberOfParticles()
        for s in self.Stacks:
           description += str(s)
        return description
        
    def AllStacks(self):
        for s in self.Stacks:
            yield s
    
    def AllParticles(self):
        for s in self.Stacks:
            for p in s.Particles:
                yield p

    def NumberOfStacks(self):
        return len(self.Stacks)

    def NumberOfParticles(self):
        Ntot = 0
        for s in self.Stacks:
            Ntot += s.NumberOfParticles()
        return Ntot

    def ReadFile(self, param_file):
        if not param_file: return
        lines = file(param_file).readlines()
        i = 0
        while True:
            i = self.GetStack(lines,i)
            if not i: break

    def GetStack(self, lines, i=0):
        if not lines: return 0
        data_block = []
        headerFound = False
        for j,l in enumerate(lines[i:]):
            line = l.strip()
            if line and not line.startswith('#'):
                if any(x in line for x in ['.mrc','.pif']):
                    if not headerFound:
                        headerFound = True
                    else:
                        break
                if headerFound: data_block.append(line)
        if data_block:
            self.Stacks.append(Stack(data_block))
        else: return 0
        
        return i+j

    def AddStack(self,st):
        self.Stacks.append(st)

    def GetDatFormat(self):
        df = []
        for st in self.Stacks:
            df.extend(st.GetDatFormat())
        return df
        
    def Write(self, outfile = ''):
        if not outfile:
            outfile = self.paramFilename
        if not outfile:
            print("ERROR(ParamFile.Write): output file name not given!")
            return
        
        df = self.GetDatFormat()
        opf = file(outfile,'w')
        opf.writelines(df)
        opf.close()
         
                
class Stack:

    def __init__(self, data_input=None):
        self.stackFilename = ''
        self.Ctf = None
        self.pixelSixe = -1.  
        self.Particles = []
        if isinstance(data_input, list):
            self.ReadBlock(data_input)
        else:
            print("Stack: can only read info from data blocks loaded as list of strings!")

    def __str__(self):
        description = "\nStack file: %s\n"%self.stackFilename
        description += " Pixel size (A): %f\n"%self.pixelSize
        description += " Voltage (kV): %d\n"%self.voltage
        description += " Cs (mm): %f\n"%self.cs
        description += " Amplitude Contrast: %f\n"%self.ampContrast
        description += str(self.Ctf)
        description += " Number of particles: %d\n"%self.NumberOfParticles()
        return description
        
    def NumberOfParticles(self):
        return len(self.Particles)

    def AllParticles(self):
        for p in self.Particles:
            yield p

    def ReadBlock(self,data_block):
        self.stackFilename = data_block[0].rsplit()[0]
        words = data_block[1].split(',')
        self.pixelSize = float(words[0])
        if int(words[1])==2: self.pixelSize *= 10.
        self.voltage = float(words[2])/1000.
        self.ampContrast = float(words[3])
        defocus1 = float(words[4])
        defocus2 = float(words[5])
        defocusAngle = float(words[6])
        self.Ctf = Ctf(defocus1, defocus2, defocusAngle)
        self.cs = float(words[7])
        if len(data_block)<3:
            return
        for line in data_block[2:]:
            if not line.startswith('#'): self.Particles.append(Particle(line))
            
    def GetDatFormat(self):
        df = []
        df.append("%s\n"%self.stackFilename)
        df.append(" %8.4f, 1,%10.1f,%10.4f,%10.4f,%10.4f,%10.4f,%8.2f\n"%(self.pixelSize,self.voltage*1000.,self.ampContrast,self.Ctf.defocusAxis1,self.Ctf.defocusAxis2,self.Ctf.defocusAngle,self.cs))
        
        for pt in self.Particles:
            df.extend(pt.GetDatFormat())
        return df
        
    def GetAvgFom(self):
        fom_avg = 0.
        count = 0
        if self.Particles:
           for p in self.Particles:
               if p.fom1:
                   if p.fom1 < 9999.:
                       fom_avg += p.fom1
                       count += 1
        if count >0: fom_avg /= count
        return fom_avg
        

class Ctf:

    def __init__(self, df1, df2, df_ang):
        self.defocusAxis1 = df1
        self.defocusAxis2 = df2
        self.defocusAngle = df_ang
        self.defocusAvg = (df1+df1)/2.
        defocusMax = max(df1,df2)
        defocusMin = min(df1,df2)
        if defocusMin>0.: self.defocusAstigRatio = defocusMax/defocusMin
        else: self.defocusAstigRatio = 0.        
        return

    def __str__(self):
        return(" Defocus axis1 (um) / axis2 (um) / angle (degs): %f / %f / %f\n"%(self.defocusAxis1,self.defocusAxis2,self.defocusAngle))


class Particle:

    def __init__(self, line=[]):
        self.id = -1
        self.theta = 0.
        self.phi = 0.
        self.omega = 0.
        self.ox = -1.
        self.oy = -1.
        self.mag = 1.
        self.fom1 = None
        self.fom2 = 0.
        self.fom3 = 0.
        self.Ctf = None
        self.flagOrient = None
        self.flagOrigin = None
        
        if line: self.read(line)
        
    def read(self, line):
        args = line.split(',')
        self.id = int(args[0])
        self.theta = float(args[1])
        self.phi = float(args[2])
        self.omega = float(args[3])
        self.ox = float(args[4]) 
        self.oy = float(args[5])
        self.mag = float(args[6]) 
        
        if len(args)>7:
            self.fom1 = float(args[7])
            self.fom2 = float(args[8])
            self.fom3 = float(args[9])
            i = 10
            if len(args) == 15:
            # from por, with ctf
                defocus1 = float(args[10])
                defocus2 = float(args[11])
                defocusAngle = float(args[12])
                self.Ctf = Ctf(defocus1, defocus2, defocusAngle)
                i = 13
            if len(args) in (12, 15):
            # from por, with ctf
                self.flagOrient = args[i].strip()
                self.flagOrigin = args[i+1].strip()

        # generate additional angles and views from euler
        # to check if conversion is correct
        psi = (-self.omega - 90.)%360.
        view_x = cos(radians(self.phi)) * sin(radians(self.theta))
        view_y = sin(radians(self.phi)) * sin(radians(self.theta))
        view_z = cos(radians(self.theta))
        angle = (psi + self.phi)%360.
        self.view = (view_x, view_y, view_z, angle)

        return

    def GetDatFormat(self):
        df = "%7d,%8.3f,%8.3f,%8.3f,%8.2f,%8.2f,%6.3f"%(self.id,self.theta,self.phi,self.omega,self.ox,self.oy,self.mag)
        if self.fom1:
            df += ",%9.3f,%9.3f,%9.3f"%(self.fom1,self.fom2,self.fom3)
        if self.Ctf:
            df += ",%7.4f,%7.4f,%7.2f"%(self.Ctf.defocusAxis1,self.Ctf.defocusAxis2,self.Ctf.defocusAngle)
        if self.flagOrient:
            df += ", %s, %s"%(self.flagOrient,self.flagOrigin)
        df += "\n"

        return [df]


if __name__== "__main__":

    import sys

    if len(sys.argv)<2:
        print("input dat file required")
        sys.exit()

    print ParamFile(sys.argv[1])
    
